from __future__ import annotations

import asyncio
import inspect
import logging
import math
import os
import pathlib
import sys
import typing as t
from functools import lru_cache
from functools import partial

import anyio.to_thread
import attrs
from simple_di import Provide
from simple_di import inject
from typing_extensions import Unpack

from bentoml import Runner
from bentoml._internal.bento.bento import Bento
from bentoml._internal.bento.build_config import BentoEnvSchema
from bentoml._internal.configuration.containers import BentoMLContainer
from bentoml._internal.context import ServiceContext
from bentoml._internal.models import Model as StoredModel
from bentoml._internal.utils import deprecated
from bentoml._internal.utils import dict_filter_none
from bentoml.exceptions import BentoMLConfigException
from bentoml.exceptions import BentoMLException

from ..images import Image
from ..method import APIMethod
from ..models import BentoModel
from ..models import HuggingFaceModel
from ..models import Model
from .config import ServiceConfig as Config

logger = logging.getLogger("bentoml.serve")

T = t.TypeVar("T", bound=object)

if t.TYPE_CHECKING:
    from bentoml._internal import external_typing as ext
    from bentoml._internal.service.openapi.specification import OpenAPISpecification
    from bentoml._internal.utils.circus import Server

    from .dependency import Dependency

    P = t.ParamSpec("P")
    R = t.TypeVar("R")

    class _ServiceDecorator(t.Protocol):
        def __call__(self, inner: type[T]) -> Service[T]: ...


def with_config(
    func: t.Callable[t.Concatenate["Service[t.Any]", P], R],
) -> t.Callable[t.Concatenate["Service[t.Any]", P], R]:
    def wrapper(self: Service[t.Any], *args: P.args, **kwargs: P.kwargs) -> R:
        self.inject_config()
        return func(self, *args, **kwargs)

    return wrapper


def convert_envs(envs: t.List[t.Dict[str, t.Any]]) -> t.List[BentoEnvSchema]:
    return [BentoEnvSchema(**env) for env in envs]


@attrs.define
class Service(t.Generic[T]):
    """A Bentoml service that can be served by BentoML server."""

    config: Config
    inner: type[T]
    image: t.Optional[Image] = None
    envs: t.List[BentoEnvSchema] = attrs.field(factory=list, converter=convert_envs)
    labels: t.Dict[str, str] = attrs.field(factory=dict)
    bento: t.Optional[Bento] = attrs.field(init=False, default=None)
    models: list[Model[t.Any]] = attrs.field(factory=list)
    apis: dict[str, APIMethod[..., t.Any]] = attrs.field(factory=dict)
    dependencies: dict[str, Dependency[t.Any]] = attrs.field(factory=dict, init=False)
    mount_apps: list[tuple[ext.ASGIApp, str, str]] = attrs.field(
        factory=list, init=False
    )
    middlewares: list[tuple[type[ext.AsgiMiddleware], dict[str, t.Any]]] = attrs.field(
        factory=list, init=False
    )
    # service context
    context: ServiceContext = attrs.field(init=False, factory=ServiceContext)
    working_dir: str = attrs.field(init=False, factory=os.getcwd)
    # import info
    _caller_module: str = attrs.field(init=False)
    _import_str: str | None = attrs.field(init=False, default=None)

    def __attrs_post_init__(self) -> None:
        from .dependency import Dependency

        has_task = False
        for field in dir(self.inner):
            value = getattr(self.inner, field)
            if isinstance(value, Dependency):
                self.dependencies[field] = value
            elif isinstance(value, StoredModel):
                logger.warning(
                    "`bentoml.models.get()` as the class attribute is not recommended because it requires the model"
                    f" to exist at import time. Use `{value._attr} = BentoModel({str(value.tag)!r})` instead."
                )
                self.models.append(BentoModel(value.tag))
            elif isinstance(value, Model):
                self.models.append(t.cast(Model[t.Any], value))
            elif isinstance(value, APIMethod):
                if value.is_task:
                    has_task = True
                self.apis[field] = t.cast("APIMethod[..., t.Any]", value)

        if has_task:
            traffic = self.config.setdefault("traffic", {})
            traffic["external_queue"] = True
            traffic.setdefault("concurrency", 1)

        pre_mount_apps = getattr(self.inner, "__bentoml_mounted_apps__", [])
        if pre_mount_apps:
            self.mount_apps.extend(pre_mount_apps)
            delattr(self.inner, "__bentoml_mounted_apps__")

    def __hash__(self):
        return hash(self.name)

    @_caller_module.default  # type: ignore
    def _get_caller_module(self) -> str:
        if __name__ == "__main__":
            return __name__
        current_frame = inspect.currentframe()
        frame = current_frame
        while frame:
            this_name = frame.f_globals["__name__"]
            if this_name != __name__:
                return this_name
            frame = frame.f_back
        return __name__

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} name={self.name!r}>"

    @lru_cache
    def find_dependent_by_path(self, path: str) -> Service[t.Any]:
        """Find a service by path"""
        attr_name, _, path = path.partition(".")
        if attr_name not in self.dependencies:
            if attr_name in self.all_services():
                return self.all_services()[attr_name]
            else:
                raise BentoMLException(f"Service {attr_name} not found")
        dependent = self.dependencies[attr_name]
        if dependent.on is None:
            raise BentoMLException(f"Service {attr_name} not found")
        if path:
            return dependent.on.find_dependent_by_path(path)
        return dependent

    def find_dependent_by_name(self, name: str) -> Service[t.Any]:
        """Find a service by name"""
        try:
            return self.all_services()[name]
        except KeyError:
            raise BentoMLException(f"Service {name} not found") from None

    @property
    def url(self) -> str | None:
        """Get the URL of the service, or None if the service is not served"""
        dependency_map = BentoMLContainer.remote_runner_mapping.get()
        url = dependency_map.get(self.name)
        return url.replace("tcp://", "http://") if url else None

    @lru_cache(maxsize=1)
    def all_services(self) -> dict[str, Service[t.Any]]:
        """Get a map of the service and all recursive dependencies"""
        services: dict[str, Service[t.Any]] = {self.name: self}
        for dependency in self.dependencies.values():
            if dependency.on is None:
                continue
            dependents = dependency.on.all_services()
            conflict = next(
                (
                    k
                    for k in dependents
                    if k in services and dependents[k] is not services[k]
                ),
                None,
            )
            if conflict:
                raise BentoMLConfigException(
                    f"Dependency conflict: {conflict} is already defined by {services[conflict].inner}"
                )
            services.update(dependents)
        return services

    @property
    def doc(self) -> str:
        from bentoml._internal.bento.bento import get_default_svc_readme

        if self.bento is not None:
            return self.bento.doc

        return get_default_svc_readme(self)

    def schema(self) -> dict[str, t.Any]:
        return dict_filter_none(
            {
                "name": self.name,
                "type": "service",
                "routes": [method.schema() for method in self.apis.values()],
                "description": getattr(self.inner, "__doc__", None),
            }
        )

    @property
    def name(self) -> str:
        name = self.config.get("name") or self.inner.__name__
        return name

    @property
    def import_string(self) -> str:
        if self._import_str is None:
            import_module = self._caller_module
            if import_module == "__main__":
                if hasattr(sys.modules["__main__"], "__file__"):
                    import_module = sys.modules["__main__"].__file__
                    assert isinstance(import_module, str)
                    try:
                        import_module_path = pathlib.Path(import_module).relative_to(
                            self.working_dir
                        )
                    except ValueError:
                        raise BentoMLException(
                            "Failed to get service import origin, service object defined in __main__ module is not supported"
                        )
                    import_module = str(import_module_path.with_suffix("")).replace(
                        os.path.sep, "."
                    )
                else:
                    raise BentoMLException(
                        "Failed to get service import origin, service object defined interactively in console or notebook is not supported"
                    )

            if self._caller_module not in sys.modules:
                raise BentoMLException(
                    "Failed to get service import origin, service object must be defined in a module"
                )

            for name, value in vars(sys.modules[self._caller_module]).items():
                if value is self:
                    self._import_str = f"{import_module}:{name}"
                    break
            else:
                raise BentoMLException(
                    "Failed to get service import origin, service object must be assigned to a variable at module level"
                )
        return self._import_str

    def to_asgi(self, is_main: bool = True) -> ext.ASGIApp:
        from _bentoml_impl.server.app import ServiceAppFactory

        self.inject_config()
        factory = ServiceAppFactory(self, is_main=is_main)
        return factory()

    def mount_asgi_app(
        self, app: ext.ASGIApp, path: str = "/", name: str | None = None
    ) -> None:
        self.mount_apps.append((app, path, name))  # type: ignore

    def mount_wsgi_app(
        self, app: ext.WSGIApp, path: str = "/", name: str | None = None
    ) -> None:
        from a2wsgi import WSGIMiddleware

        self.mount_apps.append((WSGIMiddleware(app), path, name))  # type: ignore

    def add_asgi_middleware(
        self, middleware_cls: type[ext.AsgiMiddleware], **options: t.Any
    ) -> None:
        self.middlewares.append((middleware_cls, options))

    def gradio_app_startup_hook(self, max_concurrency: int):
        gradio_apps = getattr(self.inner, "__bentoml_gradio_apps__", [])
        if gradio_apps:
            for gradio_app, path, _ in gradio_apps:
                logger.info(f"Initializing gradio app at: {path or '/'}")
                blocks = gradio_app.get_blocks()
                blocks.queue(default_concurrency_limit=max_concurrency)
                if hasattr(blocks, "startup_events"):
                    # gradio < 5.0
                    blocks.startup_events()
                else:
                    # gradio >= 5.0
                    blocks.run_startup_events()
            delattr(self.inner, "__bentoml_gradio_apps__")

    def __call__(self) -> T:
        try:
            instance = self.inner()
            instance.to_async = _AsyncWrapper(instance, self.apis.keys())
            instance.to_sync = _SyncWrapper(instance, self.apis.keys())
            return instance
        except Exception:
            logger.exception("Initializing service error")
            raise

    @property
    def openapi_spec(self) -> OpenAPISpecification:
        from .openapi import generate_spec

        return generate_spec(self)

    def inject_config(self) -> None:
        from bentoml._internal.configuration import load_config
        from bentoml._internal.configuration.containers import BentoMLContainer
        from bentoml._internal.utils import deep_merge

        # XXX: ensure at least one item to make `flatten_dict` work
        override_defaults = {
            "services": {
                name: (svc.config or {"workers": 1})
                for name, svc in self.all_services().items()
            }
        }

        load_config(override_defaults=override_defaults, use_version=2)
        main_config = BentoMLContainer.config.services[self.name].get()
        api_server_keys = (
            "traffic",
            "metrics",
            "logging",
            "ssl",
            "http",
            "grpc",
            "backlog",
            "runner_probe",
            "max_runner_connections",
        )
        api_server_config = {
            k: main_config[k] for k in api_server_keys if main_config.get(k) is not None
        }
        rest_config = {
            k: main_config[k] for k in main_config if k not in api_server_keys
        }
        existing = t.cast(t.Dict[str, t.Any], BentoMLContainer.config.get())
        deep_merge(existing, {"api_server": api_server_config, **rest_config})
        BentoMLContainer.config.set(existing)  # type: ignore

    @with_config
    @inject
    def serve_http(
        self,
        *,
        working_dir: str | None = None,
        port: int = Provide[BentoMLContainer.http.port],
        host: str = Provide[BentoMLContainer.http.host],
        backlog: int = Provide[BentoMLContainer.api_server_config.backlog],
        timeout: int | None = None,
        ssl_certfile: str | None = Provide[BentoMLContainer.ssl.certfile],
        ssl_keyfile: str | None = Provide[BentoMLContainer.ssl.keyfile],
        ssl_keyfile_password: str | None = Provide[
            BentoMLContainer.ssl.keyfile_password
        ],
        ssl_version: int | None = Provide[BentoMLContainer.ssl.version],
        ssl_cert_reqs: int | None = Provide[BentoMLContainer.ssl.cert_reqs],
        ssl_ca_certs: str | None = Provide[BentoMLContainer.ssl.ca_certs],
        ssl_ciphers: str | None = Provide[BentoMLContainer.ssl.ciphers],
        bentoml_home: str = Provide[BentoMLContainer.bentoml_home],
        development_mode: bool = False,
        reload: bool = False,
        threaded: bool = False,
    ) -> Server:
        from _bentoml_impl.server import serve_http
        from bentoml._internal.log import configure_logging

        configure_logging()

        return serve_http(
            self,
            working_dir=working_dir,
            host=host,
            port=port,
            backlog=backlog,
            timeout=timeout,
            ssl_certfile=ssl_certfile,
            ssl_keyfile=ssl_keyfile,
            ssl_keyfile_password=ssl_keyfile_password,
            ssl_version=ssl_version,
            ssl_cert_reqs=ssl_cert_reqs,
            ssl_ca_certs=ssl_ca_certs,
            ssl_ciphers=ssl_ciphers,
            bentoml_home=bentoml_home,
            development_mode=development_mode,
            reload=reload,
            threaded=threaded,
        )

    def on_load_bento(self, bento: Bento) -> None:
        service_info = next(svc for svc in bento.info.services if svc.name == self.name)
        for model, info in zip(self.models, service_info.models):
            # Replace the model version with the one in the Bento
            if not isinstance(model, HuggingFaceModel):
                continue
            model_id = info.metadata.get("model_id")  # use the case in bento info
            if not model_id:
                model_id = info.tag.name.replace("--", "/")
            revision = info.metadata.get("revision", info.tag.version)
            model.model_id = model_id
            model.revision = revision
        self.bento = bento


@t.overload
def service(inner: type[T], /) -> Service[T]: ...


@t.overload
def service(
    inner: None = ...,
    /,
    *,
    image: Image | None = None,
    envs: list[dict[str, str]] | None = None,
    labels: dict[str, str] | None = None,
    **kwargs: Unpack[Config],
) -> _ServiceDecorator: ...


def service(
    inner: type[T] | None = None,
    /,
    *,
    image: Image | None = None,
    envs: list[dict[str, str]] | None = None,
    labels: dict[str, str] | None = None,
    **kwargs: Unpack[Config],
) -> t.Any:
    """Mark a class as a BentoML service.

    Example:

        @service(traffic={"timeout": 60})
        class InferenceService:
            @api
            def predict(self, input: str) -> str:
                return input
    """
    config = kwargs

    def decorator(inner: type[T]) -> Service[T]:
        if isinstance(inner, Service):
            raise TypeError("service() decorator can only be applied once")
        return Service(
            config=config,
            inner=inner,
            image=image,
            envs=envs or [],
            labels=labels or {},
        )

    return decorator(inner) if inner is not None else decorator


@deprecated()
def runner_service(runner: Runner, **kwargs: Unpack[Config]) -> Service[t.Any]:
    """Make a service from a legacy Runner"""
    if not isinstance(runner, Runner):  # type: ignore
        raise ValueError(f"Expect an instance of Runner, but got {type(runner)}")

    class RunnerHandle(runner.runnable_class):
        def __init__(self) -> None:
            super().__init__(**runner.runnable_init_params)

    RunnerHandle.__name__ = runner.name
    apis: dict[str, APIMethod[..., t.Any]] = {}
    assert runner.runnable_class.bentoml_runnable_methods__ is not None
    for method in runner.runner_methods:
        runnable_method = runner.runnable_class.bentoml_runnable_methods__[method.name]
        api = APIMethod(  # type: ignore
            func=runnable_method.func,
            batchable=runnable_method.config.batchable,
            batch_dim=runnable_method.config.batch_dim,
            max_batch_size=method.max_batch_size,
            max_latency_ms=method.max_latency_ms,
        )
        apis[method.name] = api
    config: Config = {}
    resource_config = runner.resource_config or {}
    if (
        "nvidia.com/gpu" in runner.runnable_class.SUPPORTED_RESOURCES
        and "nvidia.com/gpu" in resource_config
    ):
        gpus: list[int] | str | int = resource_config["nvidia.com/gpu"]
        if isinstance(gpus, str):
            gpus = int(gpus)
        if runner.workers_per_resource > 1:
            config["workers"] = {}
            workers_per_resource = int(runner.workers_per_resource)
            if isinstance(gpus, int):
                gpus = list(range(gpus))
            for i in gpus:
                config["workers"].extend([{"gpus": i}] * workers_per_resource)
        else:
            resources_per_worker = int(1 / runner.workers_per_resource)
            if isinstance(gpus, int):
                config["workers"] = [
                    {"gpus": resources_per_worker}
                    for _ in range(gpus // resources_per_worker)
                ]
            else:
                config["workers"] = [
                    {"gpus": gpus[i : i + resources_per_worker]}
                    for i in range(0, len(gpus), resources_per_worker)
                ]
    elif "cpus" in resource_config:
        config["workers"] = (
            math.ceil(resource_config["cpus"]) * runner.workers_per_resource
        )
    config.update(kwargs)
    return Service(
        config=config,
        inner=RunnerHandle,
        models=[BentoModel(m.tag) for m in runner.models],
        apis=apis,
    )


class _Wrapper:
    def __init__(self, wrapped: t.Any, apis: t.Iterable[str]) -> None:
        self.__call = None
        for name in apis:
            if name == "__call__":
                self.__call = self._make_method(wrapped, name)
            else:
                setattr(self, name, self._make_method(wrapped, name))

    def __call__(self, *args: t.Any, **kwargs: t.Any) -> t.Any:
        if self.__call is None:
            raise TypeError("This service is not callable.")
        return self.__call(*args, **kwargs)

    def _make_method(self, instance: t.Any, name: str) -> t.Any:
        raise NotImplementedError


class _AsyncWrapper(_Wrapper):
    def _make_method(self, instance: t.Any, name: str) -> t.Any:
        original_func = func = getattr(instance, name).local
        while hasattr(original_func, "func"):
            original_func = original_func.func
        is_async_func = (
            asyncio.iscoroutinefunction(original_func)
            or (
                callable(original_func)
                and asyncio.iscoroutinefunction(original_func.__call__)  # type: ignore
            )
            or inspect.isasyncgenfunction(original_func)
        )
        if is_async_func:
            return func

        if inspect.isgeneratorfunction(original_func):

            async def wrapped_gen(
                *args: t.Any, **kwargs: t.Any
            ) -> t.AsyncGenerator[t.Any, None]:
                gen = func(*args, **kwargs)
                next_fun = gen.__next__
                while True:
                    try:
                        yield await anyio.to_thread.run_sync(next_fun)
                    except StopIteration:
                        break
                    except RuntimeError as e:
                        if "raised StopIteration" in str(e):
                            break
                        raise

            return wrapped_gen
        else:

            async def wrapped(*args: P.args, **kwargs: P.kwargs) -> t.Any:
                return await anyio.to_thread.run_sync(partial(func, **kwargs), *args)

            return wrapped


class _SyncWrapper(_Wrapper):
    def _make_method(self, instance: t.Any, name: str) -> t.Any:
        original_func = func = getattr(instance, name).local
        while hasattr(original_func, "func"):
            original_func = original_func.func
        is_async_func = (
            asyncio.iscoroutinefunction(original_func)
            or (
                callable(original_func)
                and asyncio.iscoroutinefunction(original_func.__call__)  # type: ignore
            )
            or inspect.isasyncgenfunction(original_func)
        )
        if not is_async_func:
            return func

        if inspect.isasyncgenfunction(original_func):

            def wrapped_gen(
                *args: t.Any, **kwargs: t.Any
            ) -> t.Generator[t.Any, None, None]:
                agen = func(*args, **kwargs)
                loop = asyncio.get_event_loop()
                while True:
                    try:
                        yield loop.run_until_complete(agen.__anext__())
                    except StopAsyncIteration:
                        break

            return wrapped_gen
        else:

            def wrapped(*args: P.args, **kwargs: P.kwargs) -> t.Any:
                loop = asyncio.get_event_loop()
                return loop.run_until_complete(func(*args, **kwargs))

            return wrapped



from __future__ import annotations

import logging
import os
import subprocess
import typing as t
from abc import abstractmethod
from functools import singledispatchmethod
from itertools import chain
from queue import Queue
from threading import Thread
from typing import TYPE_CHECKING

import attr

from ...exceptions import BentoMLException
from ..utils.filesystem import resolve_user_filepath

if TYPE_CHECKING:
    from typing_extensions import Self

    from ..types import PathType

    P = t.ParamSpec("P")

    ArgType: t.TypeAlias = tuple[str, ...] | None

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    ListStr = list[str]
else:
    ListStr = list


class Arguments(ListStr):
    def __add__(self: Self, other: Arguments) -> Arguments:
        return Arguments(super().__add__(other))

    @singledispatchmethod
    def construct_args(self, args: t.Any, opt: str = ""):
        raise NotImplementedError

    @construct_args.register(type(None))
    @construct_args.register(tuple)
    @construct_args.register(list)
    def _(self, args: ArgType, opt: str = ""):
        if args is not None:
            self.extend(
                list(chain.from_iterable(map(lambda arg: (f"--{opt}", arg), args)))
            )

    @construct_args.register(type(None))
    @construct_args.register(str)
    @construct_args.register(os.PathLike)
    def _(self, args: PathType, opt: str = ""):
        if args is not None:
            if os.path.exists(str(args)):
                args = resolve_user_filepath(str(args), ctx=None)
            self.extend((f"--{opt}", str(args)))

    @construct_args.register(type(None))
    @construct_args.register(bool)
    def _(self, args: bool, opt: str = ""):
        if args:
            self.append(f"--{opt}")


@attr.define(repr=False, init=False)
class OCIBuilder:
    """
    Base class for OCI builders. This is a command class
    that can be used to construct build commands for the given OCI builder.

    To implement a new OCI builder, provide a build and health check command via
    'container.register_backend()' instead.
    """

    binary: str = attr.field()
    env: dict[str, str] = attr.field(
        default=None, converter=attr.converters.default_if_none(factory=dict)
    )
    build_cmd: t.Sequence[str] = attr.field(
        default=None, converter=attr.converters.default_if_none(Arguments(["build"]))
    )
    enable_buildkit: bool = attr.field(default=True)

    def __repr__(self):
        return f"<OCIBuilder (binary={self.binary}, enable_buildkit={self.enable_buildkit})>"

    def __init__(
        self,
        binary: str,
        enable_buildkit: bool = True,
        build_cmd: t.Sequence[str] | None = None,
        env: dict[str, str] | None = None,
        *,
        _internal: bool = False,
    ):
        """
        Initialize the OCI builder.

        Args:
            backend: The name of the OCI builder.
            env: Environment variables to be passed to the OCI builder.
            enable_buildkit: Whether to enable BuildKit support for given OCI builder.
            build_cmd: The build command to be used by the OCI builder, minus the backend name.
                       Example: 'docker buildx build' -> ['buildx', 'build']
                       If not provided, the default will be ['build']
        """
        if not _internal:
            raise BentoMLException(
                "'OCIBuilder' cannot be instantiated directly, use 'container.get_backend()' instead."
            )
        self.__attrs_init__(binary, env, build_cmd, enable_buildkit)

    @staticmethod
    def create(
        binary: str,
        build_cmd: t.Sequence[str] | None = None,
        enable_buildkit: bool = True,
        env: dict[str, str] | None = None,
        *,
        health: t.Callable[[], bool],
        construct_build_args: t.Callable[P, list[str]],
    ) -> FrozenOCIBuilder:
        """
        Create an implementation of an OCI builder backend.
        """
        impl = FrozenOCIBuilder(
            env=env,
            binary=binary,
            build_cmd=build_cmd,
            enable_buildkit=enable_buildkit,
            _internal=True,
        )
        impl.construct_build_args = construct_build_args
        impl.health = health
        return impl

    @abstractmethod
    def health(self) -> bool:
        """
        Check if the given OCI builder is healthy.
        """
        raise NotImplementedError

    @abstractmethod
    def construct_build_args(self, **kwargs: t.Any) -> t.Sequence[str]:
        """
        Build the command for the given OCI builder.
        """
        raise NotImplementedError

    def build(self, **attrs: t.Any) -> bytes | None:
        """
        Build the OCI image, and returns the output.
        """
        context_path = attrs.get("context_path", None)
        use_sudo = attrs.pop("_run_as_root", False)
        assert context_path is not None, "context_path must be provided."
        cmds = Arguments([self.binary, *self.build_cmd])
        if use_sudo:
            cmds = ["sudo"] + cmds
        cmds.extend(self.construct_build_args(**attrs))
        logger.debug(
            "%s build cmd: '%s'", self.binary.split(os.sep)[-1], " ".join(cmds)
        )

        # We will always update suprocess environment into
        # default builder environment. This ensure that we
        # respect the user's environment.
        subprocess_env = os.environ.copy()
        env = self.env.copy()
        env.update(subprocess_env)
        commands = list(map(str, cmds))

        try:
            # We need to also respect DOCKER_BUILDKIT environment here
            # to stream logs
            if not self.enable_buildkit or (
                "DOCKER_BUILDKIT" in env and env["DOCKER_BUILDKIT"] == "0"
            ):
                return self.stream_logs(commands, cwd=context_path, env=env).stdout
            else:
                return subprocess.check_output(commands, cwd=context_path, env=env)
        except subprocess.CalledProcessError as e:
            if e.stderr:
                raise BentoMLException(e.stderr.decode("utf-8")) from None
            raise BentoMLException(str(e)) from None

    def stream_logs(
        self, cmds: t.Sequence[str], *, env: dict[str, str], cwd: PathType
    ) -> subprocess.CompletedProcess[bytes]:
        proc = subprocess.Popen(
            cmds, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=env, cwd=cwd
        )
        queue: Queue[tuple[str, bytes]] = Queue()
        stderr, stdout = b"", b""
        # We will use a thread to read from the subprocess and avoid hanging from Ctrl+C
        t = Thread(target=self._enqueue_output, args=(proc.stdout, "stdout", queue))
        t.daemon = True
        t.start()
        t = Thread(target=self._enqueue_output, args=(proc.stderr, "stderr", queue))
        t.daemon = True
        t.start()
        for _ in range(2):
            for src, line in iter(queue.get, None):
                logger.info(line.decode(errors="replace").strip("\n"))
                if src == "stderr":
                    stderr += line
                else:
                    stdout += line
        exit_code = proc.wait()
        if exit_code != 0:
            raise subprocess.CalledProcessError(
                exit_code, cmds, output=stdout, stderr=stderr
            )
        return subprocess.CompletedProcess(
            proc.args, exit_code, stdout=stdout, stderr=stderr
        )

    def _enqueue_output(
        self, pipe: t.IO[bytes], pipe_name: str, queue: Queue[tuple[str, bytes] | None]
    ):
        try:
            with pipe:
                for line in iter(pipe.readline, b""):
                    queue.put((pipe_name, line))
        finally:
            queue.put(None)


FrozenOCIBuilder = type(
    "FrozenOCIBuilder",
    (OCIBuilder,),
    {"__slots__": ("construct_build_args", "health")},
)



from __future__ import annotations

import logging
import typing as t
from pathlib import Path
from threading import Event
from threading import Thread
from typing import TYPE_CHECKING

import fs
from circus.plugins import CircusPlugin
from watchfiles import watch

from ...bento.build_config import BentoBuildConfig
from ...bento.build_config import BentoPathSpec
from ...configuration import is_editable_bentoml
from ...context import server_context
from ...log import configure_server_logging
from ...utils.pkg import source_locations

if TYPE_CHECKING:
    from watchfiles.main import FileChange


logger = logging.getLogger(__name__)


class ServiceReloaderPlugin(CircusPlugin):
    name = "service_reloader"
    config: dict[str, str]

    def __init__(self, *args: t.Any, **config: t.Any):
        assert "working_dir" in config, "`working_dir` is required"

        configure_server_logging()
        server_context.service_type = "observer"

        super().__init__(*args, **config)

        # circus/plugins/__init__.py:282 -> converts all given configs to dict[str, str]
        self.name = self.config.get("name")
        self.working_dir = self.config["working_dir"]

        # a list of folders to watch for changes
        watch_dirs = [self.working_dir]
        self._specs = [(Path(self.working_dir).as_posix(), self.create_spec())]

        if is_editable_bentoml():
            # bentoml src from this __file__
            bentoml_src = Path(source_locations("bentoml")).parent
            logger.info(
                "BentoML is installed via development mode, adding source root to 'watch_dirs'."
            )
            watch_dirs.append(str(bentoml_src))
            self._specs.append(
                (
                    bentoml_src.as_posix(),
                    BentoPathSpec(
                        # only watch python files in bentoml src
                        ["*.py", "*.yaml"],
                        [],
                        bentoml_src.as_posix(),
                        recurse_ignore_filename=".gitignore",
                    ),
                )
            )

        logger.info("Watching directories: %s", watch_dirs)

        # a thread to restart circus
        self.exit_event = Event()

        self.file_changes = watch(
            *watch_dirs,
            watch_filter=None,
            yield_on_timeout=True,  # stop hanging on our tests, doesn't affect the behaviour
            stop_event=self.exit_event,
            rust_timeout=0,  # we should set timeout to zero for no timeout
        )

    def create_spec(self) -> BentoPathSpec:
        build_config = BentoBuildConfig.from_bento_dir(self.working_dir)
        return BentoPathSpec(
            build_config.include, build_config.exclude, self.working_dir
        )

    def should_include(self, path: Path) -> bool:
        # returns True if file with 'path' has changed, else False
        str_path = path.as_posix()
        for parent, spec in self._specs:
            if fs.path.isparent(parent, str_path):
                return spec.includes(fs.path.relativefrom(parent, str_path))
        return False

    def has_modification(self) -> bool:
        for changes in self.file_changes:
            uniq_paths = {Path(c[1]) for c in changes}
            filtered = [c for c in uniq_paths if self.should_include(c)]
            if filtered:
                change_type, path = self.display_path(changes)
                logger.warning("%s: %s", change_type.upper(), path)
                return True
        return False

    def look_after(self):
        if self.has_modification():
            logger.warning("Detected changes. Reloading...")
            self.call("restart", name="*")

    def handle_init(self):
        self.watcher = Thread(target=self.look_after)
        self.watcher.start()

    def handle_recv(self, data: t.Any):
        pass

    def handle_stop(self) -> None:
        self.watcher.join()
        self.exit_event.set()

    def display_path(self, change: set[FileChange]) -> tuple[str, str]:
        type_change, path = change.pop()
        try:
            return (
                type_change.raw_str(),
                f"'{Path(path).relative_to(self.working_dir)}'",
            )
        except ValueError:
            return type_change.raw_str(), f"'{path}'"



from __future__ import annotations

import logging
import os
import re
import shutil
import typing as t
from typing import TYPE_CHECKING

import psutil

from .base import Arguments

if TYPE_CHECKING:
    from ..types import PathType
    from .base import ArgType

logger = logging.getLogger(__name__)

__all__ = ["health", "construct_build_args", "BUILDKIT_SUPPORT", "find_binary"]

BUILDKIT_SUPPORT = True


def find_binary() -> str | None:
    return shutil.which("buildctl")


def buildkitd_health() -> bool:
    if shutil.which("buildkitd") or os.getenv("BUILDKIT_HOST", "") != "":
        pass
    else:
        logger.error(
            "Both buildkitd and BUILDKIT_HOST are not found. Ensure to use either of them."
        )
        tool = shutil.which("docker") or shutil.which("podman")
        if tool is not None:
            buildkitd_instructions = "\n".join(
                [
                    "To run a portable buildkitd (via container), do the following in a terminal:",
                    "    %s run -d --name buildkitd --privileged moby/buildkit:master",
                    "    export BUILDKIT_HOST=%s-container://buildkitd",
                    "Then proceed with 'containerize' again.",
                ]
            )
            logger.info(buildkitd_instructions, tool, tool.split(os.sep)[-1])
        return False

    return True


def health() -> bool:
    client = find_binary()

    if psutil.WINDOWS:
        logger.error("buildctl is NOT SUPPORTED on Windows.")
        return False

    if not buildkitd_health():
        return False

    if client is None:
        logger.warning(
            "buildctl not found. Make sure it is installed and in your PATH. See https://github.com/moby/buildkit"
        )
        return False

    return True


def parse_dict_opt(d: dict[str, str]) -> str:
    return ",".join([f"{key}={value}" for key, value in d.items()])


ctl_args = (
    "output",
    "progress",
    "trace",
    "local",
    "frontend",
    "opt",
    "no-cache",
    "export-cache",
    "import-cache",
    "secret",
    "allow",
    "ssh",
    "metadata-file",
)

is_ctl_arg = re.compile(r"^(" + "|".join(ctl_args) + ")").search


def construct_build_args(
    *,
    file: PathType | None = None,
    tag: tuple[str] | None = None,
    context_path: PathType = ".",
    output: str | dict[str, str] | ArgType = None,
    **kwargs: t.Any,
) -> Arguments:
    # NOTE: In the case of buildctl, we will have to parse the options manually from kwargs.
    cmds = Arguments()

    opt_args: dict[str, t.Any] = {}
    buildctl_args = {
        k: v for k, v in kwargs.items() if v and is_ctl_arg(k.replace("_", "-"))
    }

    local: dict[str, str] = buildctl_args.pop("local", {})
    if isinstance(local, tuple):
        local = {k: v for (k, v) in map(lambda s: tuple(s.split("=")), local)}
    if "context" in local:
        logger.warning(
            "Passing 'context' to 'local' is not supported. If you are using Python API, use 'context_path' instead."
        )
    local["context"] = str(context_path)
    if "dockerfile" in local and file is not None:
        logger.warning(
            "Passing 'dockerfile' to 'local' is not supported. If you are using Python API, use 'file' instead."
        )
    local["dockerfile"] = os.path.dirname(str(file))
    cmds.construct_args(tuple(f"{k}={v}" for k, v in local.items()), opt="local")

    frontend: str = buildctl_args.pop("frontend", "dockerfile.v0")
    if isinstance(frontend, tuple):
        if len(frontend) > 1:
            logger.warning(
                "Multiple frontend options are not supported. Using the first one."
            )
        frontend = frontend[0]
    cmds.construct_args(frontend, opt="frontend")
    if frontend == "gateway.v0":
        opt_args["source"] = "docker/dockerfile"

    if isinstance(output, dict):
        output = parse_dict_opt(output)
    cmds.construct_args(output, opt="output")

    # handle buildctl arguments
    for k, v in buildctl_args.items():
        cmds.construct_args(v, opt=k.replace("_", "-"))

    # handle --opt
    opt_args.update(
        {k: v for k, v in kwargs.items() if is_ctl_arg(k.replace("_", "-")) is None}
    )
    for k, v in opt_args.items():
        if isinstance(v, tuple):
            cmds.construct_args(tuple(map(lambda it: f"{k}:{it}", v)), opt="opt")
        elif isinstance(v, str):
            cmds.construct_args(f"{k}={v}", opt="opt")
        else:
            raise ValueError(f"Unsupported type for {k}: {type(v)}")

    if tag is not None:
        if output is None:
            logger.warning(
                "Autoconfig for output type is deprecated and will be removed in the next major release. See message below."
            )
            # NOTE: We will always use the docker image spec if docker is available.
            # Otherwise fallback to the OCI image spec.
            if shutil.which("docker") is not None:
                cmds.construct_args(
                    tuple(map(lambda tg: f"type=docker,name=docker.io/{tg}", tag)),
                    opt="output",
                )
            else:
                cmds.construct_args(
                    tuple(map(lambda tg: f"type=oci,name={tg}", tag)),
                    opt="output",
                )
    else:
        logger.info(
            "'tag' is not specified. Result image will only be saved in build cache."
        )

    return cmds



# pragma: no cover
"""
Specific types for BentoService gRPC server.
"""
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import typing as t
    from functools import partial

    import grpc
    from grpc import aio

    from bentoml.grpc.v1.service_pb2 import Request
    from bentoml.grpc.v1.service_pb2 import Response
    from bentoml.grpc.v1.service_pb2_grpc import BentoServiceServicer

    P = t.TypeVar("P")

    BentoServicerContext = aio.ServicerContext[Request, Response]

    RequestDeserializerFn = t.Callable[[Request | None], object] | None
    ResponseSerializerFn = t.Callable[[bytes], Response | None] | None

    HandlerMethod = t.Callable[[Request, BentoServicerContext], P]
    AsyncHandlerMethod = t.Callable[[Request, BentoServicerContext], t.Awaitable[P]]

    class RpcMethodHandler(
        t.NamedTuple(
            "RpcMethodHandler",
            request_streaming=bool,
            response_streaming=bool,
            request_deserializer=RequestDeserializerFn,
            response_serializer=ResponseSerializerFn,
            unary_unary=t.Optional[HandlerMethod[Response]],
            unary_stream=t.Optional[HandlerMethod[Response]],
            stream_unary=t.Optional[HandlerMethod[Response]],
            stream_stream=t.Optional[HandlerMethod[Response]],
        ),
        grpc.RpcMethodHandler,
    ):
        """An implementation of a single RPC method."""

        request_streaming: bool
        response_streaming: bool
        request_deserializer: RequestDeserializerFn
        response_serializer: ResponseSerializerFn
        unary_unary: t.Optional[HandlerMethod[Response]]
        unary_stream: t.Optional[HandlerMethod[Response]]
        stream_unary: t.Optional[HandlerMethod[Response]]
        stream_stream: t.Optional[HandlerMethod[Response]]

    class HandlerCallDetails(
        t.NamedTuple(
            "HandlerCallDetails", method=str, invocation_metadata=aio.Metadata
        ),
        grpc.HandlerCallDetails,
    ):
        """Describes an RPC that has just arrived for service.

        Attributes:
        method: The method name of the RPC.
        invocation_metadata: A sequence of metadatum, a key-value pair included in the HTTP header.
                            An example is: ``('binary-metadata-bin', b'\\x00\\xFF')``
        """

        method: str
        invocation_metadata: aio.Metadata

    # Servicer types
    ServicerImpl = t.TypeVar("ServicerImpl")
    Servicer = t.Annotated[ServicerImpl, object]
    ServicerClass = t.Type[Servicer[t.Any]]
    AddServicerFn = t.Callable[[Servicer[t.Any], aio.Server | grpc.Server], None]

    # accepted proto fields
    ProtoField = t.Annotated[
        str,
        t.Literal[
            "dataframe",
            "file",
            "json",
            "ndarray",
            "series",
            "text",
            "multipart",
            "serialized_bytes",
        ],
    ]

    Interceptors = list[
        t.Callable[[], aio.ServerInterceptor] | partial[aio.ServerInterceptor]
    ]

    __all__ = [
        "Request",
        "Response",
        "BentoServicerContext",
        "BentoServiceServicer",
        "HandlerCallDetails",
        "RpcMethodHandler",
    ]



from __future__ import annotations

import ipaddress
import logging
import os
import re
import typing as t
from functools import singledispatch
from typing import TYPE_CHECKING

import schema as s
import yaml

from ...exceptions import BentoMLConfigException
from ..utils.lazy_loader import LazyLoader

if TYPE_CHECKING:
    from types import ModuleType

logger = logging.getLogger(__name__)

TRACING_TYPE = ["zipkin", "jaeger", "otlp", "in_memory"]


def import_configuration_spec(version: int) -> ModuleType:  # pragma: no cover
    return LazyLoader(
        f"v{version}",
        globals(),
        f"bentoml._internal.configuration.v{version}",
        exc_msg=f"Configuration version {version} does not exist.",
    )


@singledispatch
def depth(_: t.Any, _level: int = 0):  # pragma: no cover
    return _level


@depth.register(dict)
def _(d: dict[str, t.Any], level: int = 0, **kw: t.Any):
    if not d:
        return level
    return max(depth(v, level + 1, **kw) for v in d.values())


def rename_fields(
    d: dict[str, t.Any],
    current: str,
    replace_with: str | None = None,
    *,
    remove_only: bool = False,
):
    # We assume that the given dictionary is already flattened.
    # This function will rename the keys in the dictionary.
    # If `replace_with` is None, then the key will be removed.
    if depth(d) != 1:
        raise ValueError(
            "Given dictionary is not flattened. Use flatten_dict first."
        ) from None
    if current in d:
        if remove_only:
            logger.warning("Field '%s' is deprecated and will be removed." % current)
            d.pop(current)
        else:
            assert replace_with, "'replace_with' must be provided."
            logger.warning(
                "Field '%s' is deprecated and has been renamed to '%s'"
                % (current, replace_with)
            )
            d[replace_with] = d.pop(current)


punctuation = r"""!"#$%&'()*+,-./:;<=>?@[\]^`{|}~"""


def flatten_dict(
    d: t.MutableMapping[str, t.Any],
    parent: str = "",
    sep: str = ".",
) -> t.Generator[tuple[str, t.Any], None, None]:
    """Flatten nested dictionary into a single level dictionary."""
    for k, v in d.items():
        k = f'"{k}"' if any(i in punctuation for i in k) else k
        nkey = parent + sep + k if parent else k
        if isinstance(v, t.MutableMapping):
            yield from flatten_dict(
                t.cast(t.MutableMapping[str, t.Any], v), parent=nkey, sep=sep
            )
        else:
            yield nkey, v


def load_config_file(path: str) -> dict[str, t.Any]:
    """Load configuration from given path."""
    if not os.path.exists(path):
        raise BentoMLConfigException(
            "Configuration file %s not found." % path
        ) from None
    with open(path, "rb") as f:
        config = yaml.safe_load(f)
    return config


def get_default_config(version: int) -> dict[str, t.Any]:
    config = load_config_file(
        os.path.join(
            os.path.dirname(__file__), f"v{version}", "default_configuration.yaml"
        )
    )
    mod = import_configuration_spec(version)
    assert hasattr(mod, "SCHEMA"), (
        "version %d does not have a validation schema" % version
    )
    try:
        mod.SCHEMA.validate(config)
    except s.SchemaError as e:
        raise BentoMLConfigException(
            "Default configuration for version %d does not conform to given schema:\n%s"
            % (version, e)
        ) from None
    return config


def validate_tracing_type(tracing_type: str) -> bool:
    return tracing_type in TRACING_TYPE


def validate_otlp_protocol(protocol: str) -> bool:
    return protocol in ["grpc", "http"]


def ensure_larger_than(target: int | float) -> t.Callable[[int | float], bool]:
    """Ensure that given value is (lower, inf]"""

    def v(value: int | float) -> bool:
        return value > target

    return v


ensure_larger_than_zero = ensure_larger_than(0)


def ensure_range(
    lower: int | float, upper: int | float
) -> t.Callable[[int | float], bool]:
    """Ensure that given value is within the range of [lower, upper]."""

    def v(value: int | float) -> bool:
        return lower <= value <= upper

    return v


def ensure_iterable_type(typ_: type) -> t.Callable[[t.MutableSequence[t.Any]], bool]:
    """Ensure that given mutable sequence has all elements of given types."""

    def v(value: t.MutableSequence[t.Any]) -> bool:
        return all(isinstance(i, typ_) for i in value)

    return v


def is_valid_ip_address(addr: str) -> bool:
    """Check if given string is a valid IP address."""
    try:
        _ = ipaddress.ip_address(addr)
        return True
    except ValueError:
        return False


_ENV_VAR_PATTERN = r"\$\{(?P<name>.+?)(?:\:-(?P<default>.+?))?\}"


def expand_env_var(value: t.Any) -> str:
    if not isinstance(value, str) or "${" not in value:
        return value

    def _replace_env(match: t.Match[str]) -> str:
        name = match.group("name")
        default = match.group("default")
        return os.getenv(name, default or "")

    return re.sub(_ENV_VAR_PATTERN, _replace_env, value)


def expand_env_var_in_values(d: t.MutableMapping[str, t.Any]) -> None:
    for k, v in d.items():
        if isinstance(v, t.MutableMapping):
            expand_env_var_in_values(v)
        elif isinstance(v, str):
            d[k] = expand_env_var(v)
        elif isinstance(v, t.Sequence):
            d[k] = [expand_env_var(i) for i in v]



from __future__ import annotations

import contextlib
import sys
import urllib.parse
import webbrowser

import click
import rich

from bentoml._internal.cloud.client import RestApiClient
from bentoml._internal.cloud.config import DEFAULT_ENDPOINT
from bentoml._internal.cloud.config import CloudClientConfig
from bentoml._internal.cloud.config import CloudClientContext
from bentoml._internal.configuration.containers import BentoMLContainer
from bentoml._internal.utils import reserve_free_port
from bentoml._internal.utils.cattr import bentoml_cattr
from bentoml.exceptions import CLIException
from bentoml.exceptions import CloudRESTApiClientError
from bentoml_cli.auth_server import AuthCallbackHttpServer
from bentoml_cli.utils import BentoMLCommandGroup


@click.group(name="cloud", cls=BentoMLCommandGroup)
def cloud_command():
    """BentoCloud Subcommands Groups."""


@cloud_command.command()
@click.option(
    "--endpoint",
    type=click.STRING,
    help="BentoCloud endpoint",
    default=DEFAULT_ENDPOINT,
    envvar="BENTO_CLOUD_API_ENDPOINT",
    show_default=True,
    show_envvar=True,
    required=False,
)
@click.option(
    "--api-token",
    type=click.STRING,
    help="BentoCloud user API token",
    envvar="BENTO_CLOUD_API_KEY",
    show_envvar=True,
    required=False,
)
def login(endpoint: str, api_token: str) -> None:  # type: ignore (not accessed)
    """Login to BentoCloud."""
    import questionary
    from rich.prompt import Confirm

    if not api_token:
        choice = questionary.select(
            message="How would you like to authenticate BentoML CLI? [Use arrows to move]",
            choices=[
                {
                    "name": "Create a new API token with a web browser",
                    "value": "create",
                },
                {"name": "Paste an existing API token", "value": "paste"},
            ],
        ).ask()

        if choice == "create":
            with contextlib.ExitStack() as port_stack:
                port = port_stack.enter_context(
                    reserve_free_port(enable_so_reuseport=True)
                )
            callback_server = AuthCallbackHttpServer(port)
            endpoint = endpoint.rstrip("/")
            baseURL = f"{endpoint}/api_tokens"
            encodedCallback = urllib.parse.quote(callback_server.callback_url)
            authURL = f"{baseURL}?callback={encodedCallback}"
            if Confirm.ask(
                f"Would you like to open [blue]{authURL}[/] in your browser?",
                default=True,
            ):
                if webbrowser.open_new_tab(authURL):
                    rich.print(
                        f":white_check_mark: Opened [blue]{authURL}[/] in your web browser."
                    )
                else:
                    rich.print(
                        f":police_car_light: Failed to open browser. Try create a new API token at {baseURL} or Open [blue]{authURL}[/] yourself"
                    )
            else:
                rich.print(
                    f":backhand_index_pointing_right: Open [blue]{authURL}[/] yourself..."
                )
            try:
                rich.print(":hourglass: Waiting for authentication...")
                code = callback_server.wait_indefinitely_for_code()
                if code is None:
                    raise ValueError(
                        "No code could be obtained from browser callback page"
                    )
                api_token = code
            except Exception:
                rich.print(":police_car_light: Error accquiring token from web browser")
                return
        elif choice == "paste":
            api_token = click.prompt(
                "? Paste your authentication token", type=str, hide_input=True
            )
    try:
        cloud_rest_client = RestApiClient(endpoint, api_token)
        user = cloud_rest_client.v1.get_current_user()

        if user is None:
            raise CLIException("current user is not found")

        org = cloud_rest_client.v1.get_current_organization()

        if org is None:
            raise CLIException("current organization is not found")

        current_context_name = CloudClientConfig.get_config().current_context_name
        cloud_context = BentoMLContainer.cloud_context.get()

        ctx = CloudClientContext(
            name=cloud_context if cloud_context is not None else current_context_name,
            endpoint=endpoint,
            api_token=api_token,
            email=user.email,
        )

        ctx.save()
        rich.print(
            f":white_check_mark: Configured BentoCloud credentials (current-context: {ctx.name})"
        )
        rich.print(
            f":white_check_mark: Logged in as [blue]{user.email}[/] at [blue]{org.name}[/] organization"
        )
    except CloudRESTApiClientError as e:
        if e.error_code == 401:
            rich.print(
                f":police_car_light: Error validating token: HTTP 401: Bad credentials ({endpoint}/api-token)",
                file=sys.stderr,
            )
        else:
            rich.print(
                f":police_car_light: Error validating token: HTTP {e.error_code}",
                file=sys.stderr,
            )


@cloud_command.command()
def current_context() -> None:  # type: ignore (not accessed)
    """Get current cloud context."""
    rich.print_json(
        data=bentoml_cattr.unstructure(CloudClientConfig.get_config().get_context())
    )


@cloud_command.command()
def list_context() -> None:  # type: ignore (not accessed)
    """List all available context."""
    config = CloudClientConfig.get_config()
    rich.print_json(data=bentoml_cattr.unstructure([i.name for i in config.contexts]))


@cloud_command.command()
@click.argument("context_name", type=click.STRING)
def update_current_context(context_name: str) -> None:  # type: ignore (not accessed)
    """Update current context"""
    ctx = CloudClientConfig.get_config().set_current_context(context_name)
    rich.print(f"Successfully switched to context: {ctx.name}")



from __future__ import annotations

import importlib.metadata
import json
import logging
import os
import typing as t
from functools import lru_cache
from warnings import warn

from packaging.version import Version

from ...exceptions import BentoMLConfigException
from ...exceptions import BentoMLException

if t.TYPE_CHECKING:
    from .containers import SerializationStrategy

# Note this file is loaded prior to logging being configured, thus logger is only
# used within functions in this file
logger = logging.getLogger(__name__)

DEBUG_ENV_VAR = "BENTOML_DEBUG"
QUIET_ENV_VAR = "BENTOML_QUIET"
VERBOSITY_ENV_VAR = "BENTOML_VERBOSITY"
CONFIG_ENV_VAR = "BENTOML_CONFIG"
CONFIG_OVERRIDE_ENV_VAR = "BENTOML_CONFIG_OPTIONS"
CONFIG_OVERRIDE_JSON_ENV_VAR = "BENTOML_CONFIG_OVERRIDES"
# https://github.com/grpc/grpc/blob/master/doc/environment_variables.md
GRPC_DEBUG_ENV_VAR = "GRPC_VERBOSITY"
# The glibc version of python:3.11-slim image
DEFAULT_LOCK_PLATFORM = "x86_64-manylinux_2_36"


def get_bentoml_version() -> str:
    try:
        # Find BentoML version managed by setuptools_scm
        from ..._version import __version__
    except ImportError:
        # Find BentoML version managed from dist metadata
        __version__ = importlib.metadata.version("bentoml")

    return __version__


BENTOML_VERSION = get_bentoml_version()


def expand_env_var(env_var: str) -> str:
    """Expands potentially nested env var by repeatedly applying `expandvars` and
    `expanduser` until interpolation stops having any effect.
    """
    while True:
        interpolated = os.path.expanduser(os.path.expandvars(str(env_var)))
        if interpolated == env_var:
            return interpolated
        else:
            env_var = interpolated


@lru_cache(maxsize=1)
def clean_bentoml_version() -> str:
    try:
        version = Version(BENTOML_VERSION).base_version
        return str(version)
    except ValueError:
        raise BentoMLException("Errors while parsing BentoML version.") from None


@lru_cache(maxsize=1)
def is_editable_bentoml() -> bool:
    """Returns true if BentoML is installed in editable mode."""

    dist = importlib.metadata.distribution("bentoml")
    direct_url_file = next(
        (f for f in (dist.files or []) if str(f).endswith("direct_url.json")), None
    )
    if direct_url_file is None:
        return False
    direct_url_data = json.loads(direct_url_file.read_text())
    return direct_url_data.get("dir_info", {}).get("editable", False)


@lru_cache(maxsize=1)
def get_bentoml_requirement() -> str | None:
    """Returns the requirement string for BentoML."""
    dist = importlib.metadata.distribution("bentoml")
    direct_url_file = next(
        (f for f in (dist.files or []) if str(f).endswith("direct_url.json")), None
    )
    if direct_url_file is None:
        return f"bentoml=={BENTOML_VERSION}"
    direct_url_data = json.loads(direct_url_file.read_text())
    if direct_url_data.get("dir_info", {}).get("editable", False):
        return None
    if vcs_info := direct_url_data.get("vcs_info", None):
        # NOTE: uv 0.5.22 installs direct_url.json without commit_id key
        # This is a workaround to handle this case and should be removed in future
        commit_id = vcs_info.get("commit_id") or vcs_info.get("requested_revision")
        res = f"bentoml @ {vcs_info['vcs']}+{direct_url_data['url']}@{commit_id}"
    else:
        res = f"bentoml @ {direct_url_data['url']}"
    if subdirectory := direct_url_data.get("subdirectory", None):
        res += f"#subdirectory={subdirectory}"
    return res


def get_bentoml_config_file_from_env() -> str | None:
    if CONFIG_ENV_VAR in os.environ:
        # User local config file for customizing bentoml
        return expand_env_var(os.environ.get(CONFIG_ENV_VAR, ""))
    return None


def get_bentoml_override_config_from_env() -> str | None:
    if CONFIG_OVERRIDE_ENV_VAR in os.environ:
        # User local config options for customizing bentoml
        warn(
            f"Overriding BentoML config with {CONFIG_OVERRIDE_ENV_VAR} is deprecated "
            f"and will be removed in future release. Please use {CONFIG_OVERRIDE_JSON_ENV_VAR} instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return os.environ.get(CONFIG_OVERRIDE_ENV_VAR, None)
    return None


def get_bentoml_override_config_json_from_env() -> dict[str, t.Any] | None:
    if CONFIG_OVERRIDE_JSON_ENV_VAR in os.environ:
        # User local config options for customizing bentoml
        values = os.environ.get(CONFIG_OVERRIDE_JSON_ENV_VAR, None)
        if values is not None and values != "":
            return json.loads(values)

    return None


def set_verbosity(verbosity: int) -> None:
    os.environ[VERBOSITY_ENV_VAR] = str(verbosity)
    if verbosity > 0:
        os.environ[DEBUG_ENV_VAR] = "true"
        os.environ[GRPC_DEBUG_ENV_VAR] = "DEBUG"
    elif verbosity < 0:
        os.environ[QUIET_ENV_VAR] = "true"
        os.environ[GRPC_DEBUG_ENV_VAR] = "NONE"


def set_debug_mode(enabled: bool = True) -> None:
    set_verbosity(1)


def set_quiet_mode(enabled: bool = True) -> None:
    set_verbosity(-1)


def get_debug_mode() -> bool:
    if VERBOSITY_ENV_VAR in os.environ:
        return int(os.environ[VERBOSITY_ENV_VAR]) > 0
    if DEBUG_ENV_VAR in os.environ:
        return os.environ[DEBUG_ENV_VAR].lower() == "true"
    return False


def get_quiet_mode() -> bool:
    if VERBOSITY_ENV_VAR in os.environ:
        return int(os.environ[VERBOSITY_ENV_VAR]) < 0
    if QUIET_ENV_VAR in os.environ:
        return os.environ[QUIET_ENV_VAR].lower() == "true"
    return False


def load_config(
    bentoml_config_file: str | None = None,
    override_defaults: dict[str, t.Any] | None = None,
    use_version: int = 1,
):
    """Load global configuration of BentoML"""

    from .containers import BentoMLConfiguration
    from .containers import BentoMLContainer

    if not bentoml_config_file:
        bentoml_config_file = get_bentoml_config_file_from_env()

    if bentoml_config_file:
        if not bentoml_config_file.endswith((".yml", ".yaml")):
            raise BentoMLConfigException(
                f"BentoML config file specified in ENV VAR does not end with either '.yaml' or '.yml': 'BENTOML_CONFIG={bentoml_config_file}'"
            ) from None
        if not os.path.isfile(bentoml_config_file):
            raise FileNotFoundError(
                f"BentoML config file specified in ENV VAR not found: 'BENTOML_CONFIG={bentoml_config_file}'"
            ) from None

    BentoMLContainer.config.set(
        BentoMLConfiguration(
            override_config_file=bentoml_config_file,
            override_config_values=get_bentoml_override_config_from_env(),
            override_defaults=override_defaults,
            override_config_json=get_bentoml_override_config_json_from_env(),
            use_version=use_version,
        ).to_dict()
    )


def save_config(config_file_handle: t.IO[t.Any]):
    import yaml

    from ..configuration.containers import BentoMLContainer

    content = yaml.safe_dump(BentoMLContainer.config)
    config_file_handle.write(content)


def set_serialization_strategy(serialization_strategy: SerializationStrategy):
    """
    Configure how bentoml.Service are serialized (pickled). Default via EXPORT_BENTO

    Args:
        serialization_strategy: One of ["EXPORT_BENTO", "LOCAL_BENTO", "REMOTE_BENTO"]
            EXPORT_BENTO: export and compress all Bento content as bytes
            LOCAL_BENTO: load Bento by tag in other processes, assuming available in local BentoStore
            REMOTE_BENTO: load Bento by tag in other processes, pull from Yatai/BentoCloud if not available locally
    """
    from ..configuration.containers import BentoMLContainer

    BentoMLContainer.serialization_strategy.set(serialization_strategy)


@lru_cache(maxsize=1)
def get_uv_command() -> list[str]:
    """
    Get the uv command
    """
    import shutil
    import subprocess
    import sys

    if uv := shutil.which("uv"):
        return [uv]

    try:
        importlib.metadata.version("uv")
    except importlib.metadata.PackageNotFoundError:
        install_cmd = [sys.executable, "-m", "pip", "install", "-q", "uv"]
        logger.info(
            "uv is not installed, installing it with: %s", " ".join(install_cmd)
        )
        try:
            subprocess.run(
                install_cmd,
                check=True,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.PIPE,
            )
        except subprocess.CalledProcessError as e:
            raise BentoMLConfigException(
                f"Failed to install uv: {e.stderr.decode()}"
            ) from None

    return [sys.executable, "-m", "uv"]



from __future__ import annotations

import typing as t

import attr

from bentoml._internal.cloud.schemas.modelschemas import DeploymentServiceConfig
from bentoml._internal.cloud.schemas.schemasv1 import BaseListSchema
from bentoml._internal.cloud.schemas.schemasv1 import BentoWithRepositorySchema
from bentoml._internal.cloud.schemas.schemasv1 import ClusterSchema
from bentoml._internal.cloud.schemas.schemasv1 import ResourceSchema
from bentoml._internal.cloud.schemas.schemasv1 import UserSchema

from .modelschemas import EnvItemSchema
from .modelschemas import LabelItemSchema


@attr.define
class DeploymentTargetSchema(ResourceSchema):
    creator: t.Optional[UserSchema]
    config: t.Optional[DeploymentConfigSchema]
    bento: t.Optional[BentoWithRepositorySchema]
    kube_resource_uid: t.Optional[str] = attr.field(default=None)
    kube_resource_version: t.Optional[str] = attr.field(default=None)


@attr.define
class DeploymentTargetListSchema(BaseListSchema):
    items: t.List[t.Optional[DeploymentTargetSchema]]


@attr.define
class DeploymentRevisionSchema(ResourceSchema):
    creator: t.Optional[UserSchema]
    status: str
    targets: t.List[t.Optional[DeploymentTargetSchema]]


@attr.define
class DeploymentRevisionListSchema(BaseListSchema):
    items: t.List[t.Optional[DeploymentRevisionSchema]]


@attr.define
class DeploymentConfigSchema:
    __omit_if_default__ = True
    __forbid_extra_keys__ = False
    access_authorization: bool = attr.field(default=False)
    envs: t.Optional[t.List[EnvItemSchema]] = attr.field(default=None)
    labels: t.Optional[t.List[LabelItemSchema]] = attr.field(default=None)
    secrets: t.Optional[t.List[str]] = attr.field(default=None)
    services: t.Dict[str, DeploymentServiceConfig] = attr.field(factory=dict)


@attr.define(kw_only=True)
class UpdateDeploymentSchema(DeploymentConfigSchema):
    __omit_if_default__ = True
    __forbid_extra_keys__ = False  # distributed, cluster and name need to be ignored
    bento: str


@attr.define(kw_only=True)
class CreateDeploymentSchema(UpdateDeploymentSchema):
    __omit_if_default__ = True
    __forbid_extra_keys__ = False
    name: t.Optional[str] = None
    dev: bool = False


@attr.define
class DeploymentManifestSchema:
    __forbid_extra_keys__ = False
    dev: bool = False


@attr.define
class DeploymentSchema(ResourceSchema):
    __omit_if_default__ = True
    __forbid_extra_keys__ = False
    status: str
    kube_namespace: str
    creator: UserSchema
    cluster: ClusterSchema
    latest_revision: t.Optional[DeploymentRevisionSchema]
    manifest: t.Optional[DeploymentManifestSchema] = None


@attr.define
class DeploymentFullSchema(DeploymentSchema):
    urls: t.List[str] = attr.field(factory=list)


@attr.define
class DeploymentListSchema(BaseListSchema):
    __omit_if_default__ = True
    __forbid_extra_keys__ = False
    items: t.List[DeploymentSchema]


@attr.define
class KubePodStatusSchema:
    __forbid_extra_keys__ = False
    status: str
    reason: str


@attr.define
class PodStatusSchema:
    __forbid_extra_keys__ = False
    phase: str
    ready: bool


@attr.define
class KubePodSchema:
    __forbid_extra_keys__ = False
    name: str
    namespace: str
    labels: t.Dict[str, str]
    pod_status: KubePodStatusSchema
    status: PodStatusSchema
    runner_name: str


@attr.define
class LogSchema:
    __forbid_extra_keys__ = False
    items: t.List[str] = attr.field(factory=list)
    type: str = "append"


@attr.define
class LogWSResponseSchema:
    __forbid_extra_keys__ = False
    message: t.Optional[str]
    type: str
    payload: t.Optional[LogSchema]


@attr.define
class KubePodWSResponseSchema:
    __forbid_extra_keys__ = False
    message: str
    type: str
    payload: t.Optional[t.List[KubePodSchema]]


@attr.define
class UploadDeploymentFileSchema:
    __forbid_extra_keys__ = False
    path: str
    b64_encoded_content: str


@attr.define
class UploadDeploymentFilesSchema:
    __forbid_extra_keys__ = False
    files: t.List[UploadDeploymentFileSchema]


@attr.define
class DeleteDeploymentFilesSchema:
    __forbid_extra_keys__ = False
    paths: t.List[str]


@attr.define
class DeploymentFileSchema:
    __forbid_extra_keys__ = False
    path: str
    size: int
    md5: str


@attr.define
class DeploymentFileListSchema:
    __forbid_extra_keys__ = False
    files: t.List[DeploymentFileSchema]



from __future__ import annotations
from bentoml._internal.utils.pkg import pkg_version_info
from typing_extensions import deprecated
from typing import TypeVar, Callable, TYPE_CHECKING

if TYPE_CHECKING:
    from bentoml._internal.external_typing import ASGIApp

if (ver := pkg_version_info("pydantic")) < (2,):
    raise ImportError(
        f"The new SDK runs on pydantic>=2.0.0, but the you have {'.'.join(map(str, ver))}. "
        "Please upgrade it."
    )

# ruff: noqa

from .decorators import api, on_shutdown, on_startup, asgi_app, on_deployment, task
from .service import get_current_service
from .service import depends
from .service import Service, ServiceConfig
from .service import service
from .service import runner_service
from .io_models import IODescriptor

T = TypeVar("T")


@deprecated("Deprecated in favor of `bentoml.asgi_app`")
def mount_asgi_app(
    app: ASGIApp, *, path: str = "/", name: str | None = None
) -> Callable[[T], T]:
    return asgi_app(app, path=path, name=name)


__all__ = [
    "api",
    "task",
    "on_shutdown",
    "on_startup",
    "on_deployment",
    "asgi_app",
    "mount_asgi_app",
    "depends",
    "Service",
    "ServiceConfig",
    "service",
    "runner_service",
    # io descriptors
    "IODescriptor",
    "get_current_service",
]



import importlib.abc
import importlib.machinery
import logging
import os
import subprocess
import sys
import types

logger = logging.getLogger("bentoml.worker.service")


# Custom loader that calls the callback
class SafeTensorLoader(importlib.abc.Loader):
    def __init__(self, original_loader: importlib.abc.Loader):
        self.original_loader = original_loader

    def create_module(self, spec: importlib.machinery.ModuleSpec):
        return self.original_loader.create_module(spec)

    def exec_module(self, module: types.ModuleType):
        self.original_loader.exec_module(module)
        # Call the callback after the module is executed
        _do_patch(module)


# Custom finder that uses the custom loader
class SafeTensorFinder(importlib.abc.MetaPathFinder):
    def find_spec(self, fullname, path, target=None):
        if fullname == "safetensors.torch":
            for finder in sys.meta_path:
                if finder is self or not hasattr(finder, "find_spec"):
                    continue
                spec = finder.find_spec(fullname, path, target)
                if spec is not None:
                    if spec.loader is not None:
                        # Replace the loader with our custom loader
                        spec.loader = SafeTensorLoader(spec.loader)
                    return spec
        return None


def _do_patch(target_module):
    logger.info(
        "Patching safetensors.torch.safe_open to preheat model loading in parallel"
    )

    # Save the original safe_open class
    OriginalSafeOpen = target_module.safe_open

    # Define a new class to wrap around the original safe_open class
    class PatchedSafeOpen:
        def __init__(self, filename, framework, device="cpu"):
            # Call the read_ahead method before the usual safe_open
            self.read_ahead(filename)

            # Initialize the original safe_open
            self._original_safe_open = OriginalSafeOpen(filename, framework, device)

        def __enter__(self):
            return self._original_safe_open.__enter__()

        def __exit__(self, exc_type, exc_value, traceback):
            return self._original_safe_open.__exit__(exc_type, exc_value, traceback)

        @staticmethod
        def read_ahead(
            file_path,
            num_processes=None,
            size_threshold=100 * 1024 * 1024,
            block_size=1024 * 1024,
        ):
            """
            Read a file in parallel using multiple processes.

            Args:
                file_path: Path to the file to read
                num_processes: Number of processes to use for reading the file. If None, the number of processes is set to the number of CPUs.
                size_threshold: If the file size is smaller than this threshold, only one process is used to read the file.
                block_size: Block size to use for reading the file
            """
            if num_processes is None:
                num_processes = os.cpu_count() or 8

            file_size = os.path.getsize(file_path)
            if file_size <= size_threshold:
                num_processes = 1

            chunk_size = file_size // num_processes
            processes = []

            for i in range(1, num_processes):
                start_byte = i * chunk_size
                end_byte = (
                    start_byte + chunk_size if i < num_processes - 1 else file_size
                )
                logger.info(
                    f"Reading bytes {start_byte} to {end_byte} from {file_path}"
                )
                process = subprocess.Popen(
                    [
                        "dd",
                        f"if={file_path}",
                        "of=/dev/null",
                        f"bs={block_size}",
                        f"skip={start_byte // block_size}",
                        f"count={(end_byte - start_byte) // block_size}",
                        "status=none",
                    ],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                )
                processes.append(process)

        def __getattr__(self, name):
            return getattr(self._original_safe_open, name)

    # Patch the original safetensors.torch module directly
    target_module.safe_open = PatchedSafeOpen


# Insert the custom finder at the beginning of sys.meta_path
def patch_safetensor():
    sys.meta_path.insert(0, SafeTensorFinder())



from __future__ import annotations

import contextlib
import logging
import threading
import typing as t
import uuid
from queue import Empty
from urllib.parse import urlencode

import httpx
from httpx_ws import WebSocketNetworkError
from httpx_ws import connect_ws
from wsproto.utilities import LocalProtocolError

from ...exceptions import CloudRESTApiClientError
from ...exceptions import NotFound
from ..configuration import BENTOML_VERSION
from .schemas.schemasv1 import BentoListSchema
from .schemas.schemasv1 import BentoRepositorySchema
from .schemas.schemasv1 import BentoSchema
from .schemas.schemasv1 import BentoWithRepositoryListSchema
from .schemas.schemasv1 import ClusterFullSchema
from .schemas.schemasv1 import ClusterListSchema
from .schemas.schemasv1 import CompleteMultipartUploadSchema
from .schemas.schemasv1 import CreateBentoRepositorySchema
from .schemas.schemasv1 import CreateBentoSchema
from .schemas.schemasv1 import CreateDeploymentSchema as CreateDeploymentSchemaV1
from .schemas.schemasv1 import CreateModelRepositorySchema
from .schemas.schemasv1 import CreateModelSchema
from .schemas.schemasv1 import CreateSecretSchema
from .schemas.schemasv1 import DeploymentFullSchema
from .schemas.schemasv1 import DeploymentListSchema
from .schemas.schemasv1 import FinishUploadSchema
from .schemas.schemasv1 import ModelRepositorySchema
from .schemas.schemasv1 import ModelSchema
from .schemas.schemasv1 import ModelWithRepositoryListSchema
from .schemas.schemasv1 import OrganizationSchema
from .schemas.schemasv1 import PreSignMultipartUploadUrlSchema
from .schemas.schemasv1 import ResourceInstanceSchema
from .schemas.schemasv1 import SecretListSchema
from .schemas.schemasv1 import SecretSchema
from .schemas.schemasv1 import UpdateBentoSchema
from .schemas.schemasv1 import UpdateDeploymentSchema
from .schemas.schemasv1 import UpdateSecretSchema
from .schemas.schemasv1 import UserSchema
from .schemas.schemasv2 import CreateDeploymentSchema as CreateDeploymentSchemaV2
from .schemas.schemasv2 import DeleteDeploymentFilesSchema
from .schemas.schemasv2 import DeploymentFileListSchema
from .schemas.schemasv2 import DeploymentFullSchema as DeploymentFullSchemaV2
from .schemas.schemasv2 import DeploymentListSchema as DeploymentListSchemaV2
from .schemas.schemasv2 import KubePodSchema
from .schemas.schemasv2 import KubePodWSResponseSchema
from .schemas.schemasv2 import LogWSResponseSchema
from .schemas.schemasv2 import UpdateDeploymentSchema as UpdateDeploymentSchemaV2
from .schemas.schemasv2 import UploadDeploymentFilesSchema
from .schemas.utils import schema_from_object
from .schemas.utils import schema_to_object

logger = logging.getLogger(__name__)


class BaseRestApiClient:
    def __init__(self, session: httpx.Client) -> None:
        self.session = session

    @staticmethod
    def _is_not_found(resp: httpx.Response) -> bool:
        # We used to return 400 for record not found, handle both cases
        return (
            resp.status_code == 404
            or resp.status_code == 400
            and "record not found" in resp.text
        )

    @staticmethod
    def _check_resp(resp: httpx.Response) -> None:
        if resp.status_code >= 500:
            if "x-trace-id" in resp.headers:
                logger.error(
                    "request failed with status code %d: %s\n\n",
                    resp.status_code,
                    resp.text,
                )
                raise CloudRESTApiClientError(
                    f"Oops, something went wrong with BentoCloud. Please report to us with trace ID <{resp.headers['x-trace-id']}>: https://bentoml.com/support"
                )
            else:
                raise CloudRESTApiClientError(
                    f"request failed with status code {resp.status_code}: {resp.text}",
                    error_code=resp.status_code,
                )
        if resp.status_code != 200:
            raise CloudRESTApiClientError(
                f"request failed with status code {resp.status_code}: {resp.text}",
                error_code=resp.status_code,
            )


class RestApiClientV1(BaseRestApiClient):
    def get_current_user(self) -> UserSchema | None:
        url = "/api/v1/auth/current"
        resp = self.session.get(url)
        if self._is_not_found(resp):
            return None
        self._check_resp(resp)
        return schema_from_object(resp.json(), UserSchema)

    def get_current_organization(self) -> OrganizationSchema | None:
        url = "/api/v1/current_org"
        resp = self.session.get(url)
        if self._is_not_found(resp):
            return None
        self._check_resp(resp)
        return schema_from_object(resp.json(), OrganizationSchema)

    def get_bento_repository(
        self, bento_repository_name: str
    ) -> BentoRepositorySchema | None:
        url = f"/api/v1/bento_repositories/{bento_repository_name}"
        resp = self.session.get(url)
        if self._is_not_found(resp):
            return None
        self._check_resp(resp)
        return schema_from_object(resp.json(), BentoRepositorySchema)

    def create_bento_repository(
        self, req: CreateBentoRepositorySchema
    ) -> BentoRepositorySchema:
        url = "/api/v1/bento_repositories"
        resp = self.session.post(url, json=schema_to_object(req))
        self._check_resp(resp)
        return schema_from_object(resp.json(), BentoRepositorySchema)

    def get_bento(self, bento_repository_name: str, version: str) -> BentoSchema | None:
        url = f"/api/v1/bento_repositories/{bento_repository_name}/bentos/{version}"
        resp = self.session.get(url)
        if self._is_not_found(resp):
            return None
        self._check_resp(resp)
        return schema_from_object(resp.json(), BentoSchema)

    def list_bentos(
        self, bento_repository_name: str, count: int = 1
    ) -> BentoListSchema | None:
        url = f"/api/v1/bento_repositories/{bento_repository_name}/bentos"
        resp = self.session.get(url, params={"count": count})
        if self._is_not_found(resp):
            return None
        self._check_resp(resp)
        return schema_from_object(resp.json(), BentoListSchema)

    def create_bento(
        self, bento_repository_name: str, req: CreateBentoSchema
    ) -> BentoSchema:
        url = f"/api/v1/bento_repositories/{bento_repository_name}/bentos"
        resp = self.session.post(url, json=schema_to_object(req))
        self._check_resp(resp)
        return schema_from_object(resp.json(), BentoSchema)

    def update_bento(
        self, bento_repository_name: str, version: str, req: UpdateBentoSchema
    ) -> BentoSchema:
        url = f"/api/v1/bento_repositories/{bento_repository_name}/bentos/{version}"
        resp = self.session.patch(url, json=schema_to_object(req))
        self._check_resp(resp)
        return schema_from_object(resp.json(), BentoSchema)

    def presign_bento_upload_url(
        self, bento_repository_name: str, version: str
    ) -> BentoSchema:
        url = f"/api/v1/bento_repositories/{bento_repository_name}/bentos/{version}/presign_upload_url"
        resp = self.session.patch(url)
        self._check_resp(resp)
        return schema_from_object(resp.json(), BentoSchema)

    def presign_bento_download_url(
        self, bento_repository_name: str, version: str
    ) -> BentoSchema:
        url = f"/api/v1/bento_repositories/{bento_repository_name}/bentos/{version}/presign_download_url"
        resp = self.session.patch(url)
        self._check_resp(resp)
        return schema_from_object(resp.json(), BentoSchema)

    def start_bento_multipart_upload(
        self, bento_repository_name: str, version: str
    ) -> BentoSchema:
        url = f"/api/v1/bento_repositories/{bento_repository_name}/bentos/{version}/start_multipart_upload"
        resp = self.session.patch(url)
        self._check_resp(resp)
        return schema_from_object(resp.json(), BentoSchema)

    def presign_bento_multipart_upload_url(
        self,
        bento_repository_name: str,
        version: str,
        req: PreSignMultipartUploadUrlSchema,
    ) -> BentoSchema:
        url = f"/api/v1/bento_repositories/{bento_repository_name}/bentos/{version}/presign_multipart_upload_url"
        resp = self.session.patch(url, json=schema_to_object(req))
        self._check_resp(resp)
        return schema_from_object(resp.json(), BentoSchema)

    def complete_bento_multipart_upload(
        self,
        bento_repository_name: str,
        version: str,
        req: CompleteMultipartUploadSchema,
    ) -> BentoSchema:
        url = f"/api/v1/bento_repositories/{bento_repository_name}/bentos/{version}/complete_multipart_upload"
        resp = self.session.patch(url, json=schema_to_object(req))
        self._check_resp(resp)
        return schema_from_object(resp.json(), BentoSchema)

    def start_upload_bento(
        self, bento_repository_name: str, version: str
    ) -> BentoSchema:
        url = f"/api/v1/bento_repositories/{bento_repository_name}/bentos/{version}/start_upload"
        resp = self.session.patch(url)
        self._check_resp(resp)
        return schema_from_object(resp.json(), BentoSchema)

    def finish_upload_bento(
        self, bento_repository_name: str, version: str, req: FinishUploadSchema
    ) -> BentoSchema:
        url = f"/api/v1/bento_repositories/{bento_repository_name}/bentos/{version}/finish_upload"
        resp = self.session.patch(url, json=schema_to_object(req))
        self._check_resp(resp)
        return schema_from_object(resp.json(), BentoSchema)

    def upload_bento(
        self, bento_repository_name: str, version: str, data: t.IO[bytes]
    ) -> None:
        url = f"/api/v1/bento_repositories/{bento_repository_name}/bentos/{version}/upload"
        resp = self.session.put(
            url,
            content=data,
            headers=dict(
                self.session.headers, **{"Content-Type": "application/octet-stream"}
            ),
        )
        self._check_resp(resp)
        return None

    @contextlib.contextmanager
    def download_bento(
        self, bento_repository_name: str, version: str
    ) -> t.Generator[httpx.Response, None, None]:
        url = f"/api/v1/bento_repositories/{bento_repository_name}/bentos/{version}/download"
        with self.session.stream("GET", url) as resp:
            self._check_resp(resp)
            yield resp

    def get_model_repository(
        self, model_repository_name: str
    ) -> ModelRepositorySchema | None:
        url = f"/api/v1/model_repositories/{model_repository_name}"
        resp = self.session.get(url)
        if self._is_not_found(resp):
            return None
        self._check_resp(resp)
        return schema_from_object(resp.json(), ModelRepositorySchema)

    def create_model_repository(
        self, req: CreateModelRepositorySchema
    ) -> ModelRepositorySchema:
        url = "/api/v1/model_repositories"
        resp = self.session.post(url, json=schema_to_object(req))
        self._check_resp(resp)
        return schema_from_object(resp.json(), ModelRepositorySchema)

    def get_model(self, model_repository_name: str, version: str) -> ModelSchema | None:
        url = f"/api/v1/model_repositories/{model_repository_name}/models/{version}"
        resp = self.session.get(url)
        if self._is_not_found(resp):
            return None
        self._check_resp(resp)
        return schema_from_object(resp.json(), ModelSchema)

    def create_model(
        self, model_repository_name: str, req: CreateModelSchema
    ) -> ModelSchema:
        url = f"/api/v1/model_repositories/{model_repository_name}/models"
        resp = self.session.post(url, json=schema_to_object(req))
        self._check_resp(resp)
        return schema_from_object(resp.json(), ModelSchema)

    def presign_model_upload_url(
        self, model_repository_name: str, version: str
    ) -> ModelSchema:
        url = f"/api/v1/model_repositories/{model_repository_name}/models/{version}/presign_upload_url"
        resp = self.session.patch(url)
        self._check_resp(resp)
        return schema_from_object(resp.json(), ModelSchema)

    def presign_model_download_url(
        self, model_repository_name: str, version: str
    ) -> ModelSchema:
        url = f"/api/v1/model_repositories/{model_repository_name}/models/{version}/presign_download_url"
        resp = self.session.patch(url)
        self._check_resp(resp)
        return schema_from_object(resp.json(), ModelSchema)

    def start_model_multipart_upload(
        self, model_repository_name: str, version: str
    ) -> ModelSchema:
        url = f"/api/v1/model_repositories/{model_repository_name}/models/{version}/start_multipart_upload"
        resp = self.session.patch(url)
        self._check_resp(resp)
        return schema_from_object(resp.json(), ModelSchema)

    def presign_model_multipart_upload_url(
        self,
        model_repository_name: str,
        version: str,
        req: PreSignMultipartUploadUrlSchema,
    ) -> ModelSchema:
        url = f"/api/v1/model_repositories/{model_repository_name}/models/{version}/presign_multipart_upload_url"
        resp = self.session.patch(url, json=schema_to_object(req))
        self._check_resp(resp)
        return schema_from_object(resp.json(), ModelSchema)

    def complete_model_multipart_upload(
        self,
        model_repository_name: str,
        version: str,
        req: CompleteMultipartUploadSchema,
    ) -> ModelSchema:
        url = f"/api/v1/model_repositories/{model_repository_name}/models/{version}/complete_multipart_upload"
        resp = self.session.patch(url, json=schema_to_object(req))
        self._check_resp(resp)
        return schema_from_object(resp.json(), ModelSchema)

    def start_upload_model(
        self, model_repository_name: str, version: str
    ) -> ModelSchema:
        url = f"/api/v1/model_repositories/{model_repository_name}/models/{version}/start_upload"
        resp = self.session.patch(url)
        self._check_resp(resp)
        return schema_from_object(resp.json(), ModelSchema)

    def finish_upload_model(
        self, model_repository_name: str, version: str, req: FinishUploadSchema
    ) -> ModelSchema:
        url = f"/api/v1/model_repositories/{model_repository_name}/models/{version}/finish_upload"
        resp = self.session.patch(url, json=schema_to_object(req))
        self._check_resp(resp)
        return schema_from_object(resp.json(), ModelSchema)

    def upload_model(
        self, model_repository_name: str, version: str, data: t.IO[bytes]
    ) -> None:
        url = f"/api/v1/model_repositories/{model_repository_name}/models/{version}/upload"
        resp = self.session.put(
            url,
            content=data,
            headers=dict(
                self.session.headers, **{"Content-Type": "application/octet-stream"}
            ),
        )
        self._check_resp(resp)
        return None

    @contextlib.contextmanager
    def download_model(
        self, model_repository_name: str, version: str
    ) -> t.Generator[httpx.Response, None, None]:
        url = f"/api/v1/model_repositories/{model_repository_name}/models/{version}/download"
        with self.session.stream("GET", url) as resp:
            self._check_resp(resp)
            yield resp

    def get_bento_repositories_list(
        self, bento_repository_name: str
    ) -> BentoWithRepositoryListSchema | None:
        url = "/api/v1/bento_repositories"
        resp = self.session.get(url)
        if self._is_not_found(resp):
            return None
        self._check_resp(resp)
        return schema_from_object(resp.json(), BentoWithRepositoryListSchema)

    def get_bentos_list(self) -> BentoWithRepositoryListSchema | None:
        url = "/api/v1/bentos"
        resp = self.session.get(url)
        if self._is_not_found(resp):
            return None
        self._check_resp(resp)
        return schema_from_object(resp.json(), BentoWithRepositoryListSchema)

    def get_models_list(self) -> ModelWithRepositoryListSchema | None:
        url = "/api/v1/models"
        resp = self.session.get(url)
        if self._is_not_found(resp):
            return None
        self._check_resp(resp)
        return schema_from_object(resp.json(), ModelWithRepositoryListSchema)

    def get_cluster_deployment_list(
        self, cluster: str, **params: str | int | None
    ) -> DeploymentListSchema | None:
        url = f"/api/v1/clusters/{cluster}/deployments"
        resp = self.session.get(url, params=params)
        if self._is_not_found(resp):
            return None
        self._check_resp(resp)
        return schema_from_object(resp.json(), DeploymentListSchema)

    def get_organization_deployment_list(
        self, **params: str | int | None
    ) -> DeploymentListSchema | None:
        url = "/api/v1/deployments"
        resp = self.session.get(url, params=params)
        if self._is_not_found(resp):
            return None
        self._check_resp(resp)
        return schema_from_object(resp.json(), DeploymentListSchema)

    def create_deployment(
        self, cluster: str, create_schema: CreateDeploymentSchemaV1
    ) -> DeploymentFullSchema | None:
        url = f"/api/v1/clusters/{cluster}/deployments"
        resp = self.session.post(url, json=schema_to_object(create_schema))
        self._check_resp(resp)
        return schema_from_object(resp.json(), DeploymentFullSchema)

    def get_deployment(
        self, cluster: str, kube_namespace: str, name: str
    ) -> DeploymentFullSchema | None:
        url = (
            f"/api/v1/clusters/{cluster}/namespaces/{kube_namespace}/deployments/{name}"
        )
        resp = self.session.get(url)
        if self._is_not_found(resp):
            return None
        self._check_resp(resp)
        return schema_from_object(resp.json(), DeploymentFullSchema)

    def update_deployment(
        self,
        cluster: str,
        kube_namespace: str,
        name: str,
        update_schema: UpdateDeploymentSchema,
    ) -> DeploymentFullSchema | None:
        url = (
            f"/api/v1/clusters/{cluster}/namespaces/{kube_namespace}/deployments/{name}"
        )
        resp = self.session.patch(url, json=schema_to_object(update_schema))
        if self._is_not_found(resp):
            return None
        self._check_resp(resp)
        return schema_from_object(resp.json(), DeploymentFullSchema)

    def terminate_deployment(
        self, cluster: str, kube_namespace: str, name: str
    ) -> DeploymentFullSchema | None:
        url = f"/api/v1/clusters/{cluster}/namespaces/{kube_namespace}/deployments/{name}/terminate"
        resp = self.session.post(url)
        if self._is_not_found(resp):
            return None
        self._check_resp(resp)
        return schema_from_object(resp.json(), DeploymentFullSchema)

    def delete_deployment(
        self, cluster: str, kube_namespace: str, name: str
    ) -> DeploymentFullSchema | None:
        url = (
            f"/api/v1/clusters/{cluster}/namespaces/{kube_namespace}/deployments/{name}"
        )
        resp = self.session.delete(url)
        if self._is_not_found(resp):
            return None
        self._check_resp(resp)
        return schema_from_object(resp.json(), DeploymentFullSchema)

    def get_cluster_list(
        self, params: dict[str, str | int] | None = None
    ) -> ClusterListSchema | None:
        url = "/api/v1/clusters"
        resp = self.session.get(url, params=params)
        if self._is_not_found(resp):
            return None
        self._check_resp(resp)
        return schema_from_object(resp.json(), ClusterListSchema)

    def get_cluster(self, cluster: str) -> ClusterFullSchema | None:
        url = f"/api/v1/clusters/{cluster}"
        resp = self.session.get(url)
        if self._is_not_found(resp):
            return None
        self._check_resp(resp)
        return schema_from_object(resp.json(), ClusterFullSchema)

    def get_latest_model(
        self, model_repository_name: str, query: str | None = None
    ) -> ModelSchema | None:
        url = f"/api/v1/model_repositories/{model_repository_name}/models"
        params = {"start": "0", "count": "10"}
        if query:
            params["q"] = query
        resp = self.session.get(url, params=params)
        self._check_resp(resp)
        models = resp.json()["items"]
        return schema_from_object(models[0], ModelSchema) if models else None

    def list_secrets(
        self,
        count: int | None = None,
        q: str | None = None,
        search: str | None = None,
        start: int | None = None,
    ) -> SecretListSchema:
        url = "/api/v1/org_secrets"
        if not count:
            count = 10
        if not start:
            start = 0
        resp = self.session.get(
            url,
            params={
                "count": count,
                "q": q,
                "search": search,
                "start": start,
                "all": "true",
            },
        )
        self._check_resp(resp)
        return schema_from_object(resp.json(), SecretListSchema)

    def create_secret(
        self, secret: CreateSecretSchema, cluster: str | None = None
    ) -> SecretSchema:
        url = "/api/v1/org_secrets"
        if cluster:
            url = f"{url}?cluster={cluster}"
        resp = self.session.post(url, json=schema_to_object(secret))
        self._check_resp(resp)
        return schema_from_object(resp.json(), SecretSchema)

    def get_secret(self, name: str, cluster: str | None = None) -> SecretSchema:
        url = f"/api/v1/org_secrets/{name}"
        if cluster:
            url = f"{url}?cluster={cluster}"
        resp = self.session.get(url)
        self._check_resp(resp)
        return schema_from_object(resp.json(), SecretSchema)

    def delete_secret(self, name: str, cluster: str | None = None) -> None:
        url = f"/api/v1/org_secrets/{name}"
        if cluster:
            url = f"{url}?cluster={cluster}"
        resp = self.session.delete(url)
        self._check_resp(resp)

    def update_secret(
        self, name: str, secret: UpdateSecretSchema, cluster: str | None = None
    ) -> SecretSchema:
        url = f"/api/v1/org_secrets/{name}"
        if cluster:
            url = f"{url}?cluster={cluster}"
        resp = self.session.patch(url, json=schema_to_object(secret))
        self._check_resp(resp)
        return schema_from_object(resp.json(), SecretSchema)


class RestApiClientV2(BaseRestApiClient):
    def create_deployment(
        self,
        create_schema: CreateDeploymentSchemaV2,
        cluster: str | None = None,
    ) -> DeploymentFullSchemaV2:
        url = "/api/v2/deployments"
        resp = self.session.post(
            url, json=schema_to_object(create_schema), params={"cluster": cluster}
        )
        self._check_resp(resp)
        return schema_from_object(resp.json(), DeploymentFullSchemaV2)

    def update_deployment(
        self,
        name: str,
        update_schema: UpdateDeploymentSchemaV2,
        cluster: str | None = None,
    ) -> DeploymentFullSchemaV2:
        url = f"/api/v2/deployments/{name}"
        resp = self.session.put(
            url, json=schema_to_object(update_schema), params={"cluster": cluster}
        )
        if self._is_not_found(resp):
            raise NotFound(f"Deployment {name} is not found: {resp.text}")
        self._check_resp(resp)
        return schema_from_object(resp.json(), DeploymentFullSchemaV2)

    def void_update_deployment(
        self, name: str, cluster: str | None
    ) -> DeploymentFullSchemaV2:
        url = f"/api/v2/deployments/{name}/void_update"
        resp = self.session.put(url, params={"cluster": cluster})
        if self._is_not_found(resp):
            raise NotFound(f"Deployment {name} is not found: {resp.text}")
        self._check_resp(resp)
        return schema_from_object(resp.json(), DeploymentFullSchemaV2)

    def get_deployment(
        self,
        name: str,
        cluster: str | None = None,
    ) -> DeploymentFullSchemaV2:
        url = f"/api/v2/deployments/{name}"
        resp = self.session.get(url, params={"cluster": cluster})
        if self._is_not_found(resp):
            raise NotFound(f"Deployment {name} is not found: {resp.text}")
        self._check_resp(resp)
        return schema_from_object(resp.json(), DeploymentFullSchemaV2)

    def list_deployment(
        self,
        cluster: str | None = None,
        all: bool | None = None,
        # if both of the above is none, list default cluster's deployments
        count: int | None = None,
        q: str | None = None,
        search: str | None = None,
        start: int | None = None,
    ) -> DeploymentListSchemaV2:
        url = "/api/v2/deployments"
        resp = self.session.get(
            url,
            params={
                "cluster": cluster,
                "all": all,
                "count": count,
                "q": q,
                "search": search,
                "start": start,
            },
        )
        if self._is_not_found(resp):
            raise NotFound(f"Deployment not found: {resp.text}")
        self._check_resp(resp)
        return schema_from_object(resp.json(), DeploymentListSchemaV2)

    def terminate_deployment(
        self,
        name: str,
        cluster: str | None = None,
    ) -> DeploymentFullSchemaV2:
        url = f"/api/v2/deployments/{name}/terminate"
        resp = self.session.post(url, params={"cluster": cluster})
        if self._is_not_found(resp):
            raise NotFound(f"Deployment {name} is not found: {resp.text}")
        self._check_resp(resp)
        return schema_from_object(resp.json(), DeploymentFullSchemaV2)

    def delete_deployment(
        self,
        name: str,
        cluster: str | None = None,
    ) -> DeploymentFullSchemaV2:
        url = f"/api/v2/deployments/{name}"
        resp = self.session.delete(url, params={"cluster": cluster})
        if self._is_not_found(resp):
            raise NotFound(f"Deployment {name} is not found: {resp.text}")
        self._check_resp(resp)
        return schema_from_object(resp.json(), DeploymentFullSchemaV2)

    def list_instance_types(
        self,
        cluster: str | None = None,
    ) -> list[ResourceInstanceSchema]:
        url = "/api/v1/instance_types"
        resp = self.session.get(url, params={"cluster": cluster})
        self._check_resp(resp)
        return schema_from_object(resp.json(), list[ResourceInstanceSchema])

    def get_deployment_image_builder_pod(
        self, name: str, cluster: str | None = None
    ) -> KubePodSchema | None:
        deployment = self.get_deployment(name, cluster=cluster)
        if not deployment.latest_revision:
            raise NotFound(f"Deployment {name} latest revision is not found")
        if not deployment.latest_revision.targets:
            raise NotFound(f"Deployment {name} latest revision targets is not found")
        target = deployment.latest_revision.targets[0]
        if not target:
            raise NotFound(f"Deployment {name} latest revision target is not found")
        if not target.bento:
            raise NotFound(
                f"Deployment {name} latest revision target bento is not found"
            )
        scheme = "wss"
        base_url = self.session.base_url
        if base_url.scheme == "http":
            scheme = "ws"
        endpoint = f"{scheme}://{base_url.netloc.decode('utf-8')}"
        with connect_ws(
            url=f"{endpoint}/ws/v1/clusters/{deployment.cluster.name}/pods?"
            f"{urlencode(dict(organization_name=deployment.cluster.organization_name, namespace=deployment.kube_namespace, selector=f'yatai.ai/bento-repository={target.bento.repository.name},yatai.ai/bento={target.bento.version}'))}",
            client=self.session,
        ) as ws:
            jsn = schema_from_object(ws.receive_json(), KubePodWSResponseSchema)
            if jsn.type == "error":
                raise CloudRESTApiClientError(jsn.message)
            for pod in jsn.payload or []:
                if pod.labels.get("yatai.ai/is-bento-image-builder") == "true":
                    return pod
        return None

    def list_deployment_pods(
        self, name: str, cluster: str | None = None
    ) -> list[KubePodSchema]:
        deployment = self.get_deployment(name, cluster=cluster)
        if not deployment.latest_revision:
            raise NotFound(f"Deployment {name} latest revision is not found")
        if not deployment.latest_revision.targets:
            raise NotFound(f"Deployment {name} latest revision targets is not found")
        target = deployment.latest_revision.targets[0]
        if not target:
            raise NotFound(f"Deployment {name} latest revision target is not found")
        if not target.bento:
            raise NotFound(
                f"Deployment {name} latest revision target bento is not found"
            )
        scheme = "wss"
        base_url = self.session.base_url
        if base_url.scheme == "http":
            scheme = "ws"
        endpoint = f"{scheme}://{base_url.netloc.decode('utf-8')}"
        with connect_ws(
            url=f"{endpoint}/ws/v2/deployments/{name}/pods?{urlencode(dict(organization_name=deployment.cluster.organization_name, cluster=deployment.cluster.name))}",
            client=self.session,
        ) as ws:
            jsn = schema_from_object(ws.receive_json(), KubePodWSResponseSchema)
            if jsn.type == "error":
                raise CloudRESTApiClientError(jsn.message)
            return jsn.payload or []

    def tail_logs(
        self,
        *,
        cluster_name: str,
        namespace: str,
        pod_name: str,
        container_name: str = "main",
        stop_event: threading.Event,
    ) -> t.Generator[str, None, None]:
        scheme = "wss"
        base_url = self.session.base_url
        if base_url.scheme == "http":
            scheme = "ws"
        endpoint = f"{scheme}://{base_url.netloc.decode('utf-8')}"

        with connect_ws(
            url=f"{endpoint}/ws/v1/clusters/{cluster_name}/tail?{urlencode(dict(namespace=namespace, pod_name=pod_name, timestamps='false'))}",
            client=self.session,
        ) as ws:
            req_id = str(uuid.uuid4())
            ws.send_json(
                {
                    "type": "data",
                    "payload": {
                        "id": req_id,
                        "container_name": container_name,
                        "follow": True,
                        "tail_lines": 50,
                    },
                }
            )

            def heartbeat() -> None:
                while True:
                    try:
                        ws.send_json({"type": "heartbeat"})
                    except (WebSocketNetworkError, LocalProtocolError):
                        pass
                    if stop_event.wait(5):
                        break

            heartbeat_thread = threading.Thread(target=heartbeat)
            heartbeat_thread.start()

            try:
                while True:
                    if stop_event.is_set():
                        break
                    try:
                        data = ws.receive_json(timeout=1)
                    except Empty:
                        continue
                    jsn = schema_from_object(data, LogWSResponseSchema)
                    if jsn.type == "error":
                        raise CloudRESTApiClientError(jsn.message or "Unknown error")
                    if jsn.type == "success" and jsn.message == "EOF":
                        break
                    if jsn.type == "heartbeat" or jsn.payload is None:
                        continue
                    for line in jsn.payload.items:
                        yield line
            except WebSocketNetworkError:
                pass
            finally:
                stop_event.set()
                heartbeat_thread.join()

    def upload_files(
        self, name: str, files: UploadDeploymentFilesSchema, cluster: str | None = None
    ) -> None:
        url = f"/api/v2/deployments/{name}/files"
        resp = self.session.post(
            url, json=schema_to_object(files), params={"cluster": cluster}, timeout=600
        )
        self._check_resp(resp)

    def delete_files(
        self, name: str, paths: DeleteDeploymentFilesSchema, cluster: str | None = None
    ) -> None:
        url = f"/api/v2/deployments/{name}/files"
        resp = self.session.request(
            "DELETE", url, json=schema_to_object(paths), params={"cluster": cluster}
        )
        self._check_resp(resp)

    def list_files(
        self, name: str, cluster: str | None = None
    ) -> DeploymentFileListSchema:
        url = f"/api/v2/deployments/{name}/files"
        resp = self.session.get(url, params={"cluster": cluster})
        self._check_resp(resp)
        return schema_from_object(resp.json(), DeploymentFileListSchema)


class RestApiClient:
    def __init__(self, endpoint: str, api_token: str, timeout: int = 60) -> None:
        self.endpoint = endpoint
        headers = {"X-YATAI-API-TOKEN": api_token, "X-Bentoml-Version": BENTOML_VERSION}
        self.session = httpx.Client(base_url=endpoint, timeout=timeout, headers=headers)
        self.v2 = RestApiClientV2(self.session)
        self.v1 = RestApiClientV1(self.session)



"""
_bentoml_impl.server
~~~~~~~~~~~~~~~~~~~~

A reference implementation of serving a BentoML service.
This will be eventually migrated to Rust.
"""

from .serving import serve_http

__all__ = ["serve_http"]



from __future__ import annotations

import importlib
import inspect
import logging
import os
import sys
import typing as t
from functools import partial

import attr

from bentoml.exceptions import BentoMLException

from ...exceptions import NotFound
from ...grpc.utils import LATEST_PROTOCOL_VERSION
from ...grpc.utils import import_grpc
from ..bento.bento import get_default_svc_readme
from ..context import ServiceContext as Context
from ..io_descriptors import IODescriptor
from ..io_descriptors.base import IOType
from ..models import Model
from ..runner.runner import AbstractRunner
from ..runner.runner import Runner
from ..tag import Tag
from ..utils import deprecated
from ..utils import first_not_none
from .inference_api import InferenceAPI

if t.TYPE_CHECKING:
    import grpc

    import bentoml
    from bentoml.grpc.types import AddServicerFn
    from bentoml.grpc.types import ServicerClass
    from bentoml.triton import _TritonRunner

    from ...grpc.v1 import service_pb2_grpc as services
    from .. import external_typing as ext
    from ..bento import Bento
    from .openapi.specification import OpenAPISpecification

    Hook = t.Callable[[], None | t.Coroutine[t.Any, t.Any, None]]
    ContextFunc = t.Callable[[Context], None | t.Coroutine[t.Any, t.Any, None]]
    HookF = t.TypeVar("HookF", bound=Hook)
    HookF_ctx = t.TypeVar("HookF_ctx", bound=ContextFunc)

    class _inference_api_wrapper(t.Generic[IOType]):
        __name__: str

        # fmt: off
        @t.overload
        def __call__(self, func: t.Callable[[IOType], IOType]) -> t.Callable[[IOType], IOType]: ...  # type: ignore (this is considered as stub)
        @t.overload
        def __call__(self, func: t.Callable[[IOType], t.Coroutine[IOType, t.Any, t.Any]]) -> t.Callable[[IOType], t.Coroutine[IOType, t.Any, t.Any]]: ...
        @t.overload
        def __call__(self, func: t.Callable[[IOType, bentoml.Context], IOType]) -> t.Callable[[IOType, bentoml.Context], IOType]: ...
        @t.overload
        def __call__(self, func: t.Callable[[IOType, bentoml.Context], t.Coroutine[IOType, t.Any, t.Any]]) -> t.Callable[[IOType, bentoml.Context], t.Coroutine[IOType, t.Any, t.Any]]: ...
        # fmt: on

else:
    grpc, _ = import_grpc()

logger = logging.getLogger(__name__)


def get_valid_service_name(user_provided_svc_name: str) -> str:
    lower_name = user_provided_svc_name.lower()

    if user_provided_svc_name != lower_name:
        logger.warning(
            "Converting %s to lowercase: %s.", user_provided_svc_name, lower_name
        )

    # Service name must be a valid Tag name; create a dummy tag to use its validation
    Tag(lower_name)
    return lower_name


@deprecated("bentoml.Service", suggestion="Please upgrade to @bentoml.service().")
@attr.define(frozen=False, init=False)
class Service:
    """The service definition is the manifestation of the Service Oriented Architecture
    and the core building block in BentoML where users define the service runtime
    architecture and model serving logic.

    A BentoML service is defined via instantiate this Service class. When creating a
    Service instance, user must provide a Service name and list of runners that are
    required by this Service. The instance can then be used to define InferenceAPIs via
    the `api` decorator.
    """

    name: str
    runners: t.List[Runner | _TritonRunner]
    models: t.List[Model]

    # starlette related
    mount_apps: t.List[t.Tuple[ext.ASGIApp, str, str]] = attr.field(
        init=False, factory=list
    )
    middlewares: t.List[t.Tuple[t.Type[ext.AsgiMiddleware], t.Dict[str, t.Any]]] = (
        attr.field(init=False, factory=list)
    )

    # gRPC related
    mount_servicers: list[tuple[ServicerClass, AddServicerFn, list[str]]] = attr.field(
        init=False, factory=list
    )
    interceptors: list[partial[grpc.aio.ServerInterceptor]] = attr.field(
        init=False, factory=list
    )
    grpc_handlers: list[grpc.GenericRpcHandler] = attr.field(init=False, factory=list)

    # list of APIs from @svc.api
    apis: t.Dict[str, InferenceAPI[t.Any]] = attr.field(init=False, factory=dict)

    # Tag/Bento are only set when the service was loaded from a bento
    tag: Tag | None = attr.field(init=False, default=None)
    bento: Bento | None = attr.field(init=False, default=None)

    # Working dir and Import path of the service, set when the service was imported
    _working_dir: str | None = attr.field(init=False, default=None)
    _import_str: str | None = attr.field(init=False, default=None)
    _caller_module: str | None = attr.field(init=False, default=None)

    # service context
    context: Context = attr.field(init=False, factory=Context)

    # hooks
    startup_hooks: list[Hook] = attr.field(init=False, factory=list)
    shutdown_hooks: list[Hook] = attr.field(init=False, factory=list)
    deployment_hooks: list[Hook] = attr.field(init=False, factory=list)

    def __reduce__(self):
        """
        Make bentoml.Service pickle serializable
        """
        import fs

        from bentoml._internal.bento.bento import Bento
        from bentoml._internal.configuration.containers import BentoMLContainer
        from bentoml._internal.service.loader import load
        from bentoml._internal.service.loader import load_bento_dir

        serialization_strategy = BentoMLContainer.serialization_strategy.get()

        if self.bento:
            if serialization_strategy == "EXPORT_BENTO":
                temp_fs = fs.open_fs("temp:///")
                tmp_path = temp_fs.getsyspath("/")
                bento_path = self.bento.export(tmp_path, output_format="tar")
                content = open(bento_path, "rb").read()

                def load_exported_bento(bento_tag: Tag, content: bytes):
                    tmp_bento_store = BentoMLContainer.tmp_bento_store.get()
                    try:
                        bento = tmp_bento_store.get(bento_tag)
                        return load_bento_dir(bento.path)
                    except NotFound:
                        temp_fs = fs.open_fs("temp:///")
                        temp_fs.writebytes("/import.bento", content)
                        bento = Bento.import_from(
                            temp_fs.getsyspath("/import.bento")
                        ).save(tmp_bento_store)
                        return load_bento_dir(bento.path)

                return (
                    load_exported_bento,
                    (
                        self.bento.tag,
                        content,
                    ),
                )
            elif serialization_strategy == "LOCAL_BENTO":
                return (load, (self.bento.tag,))
            else:
                # serialization_strategy == REMOTE_BENTO
                def get_or_pull(bento_tag: Tag) -> Service:
                    try:
                        return load(bento_tag)
                    except NotFound:
                        pass
                    tmp_bento_store = BentoMLContainer.tmp_bento_store.get()
                    try:
                        bento = tmp_bento_store.get(bento_tag)
                        return load_bento_dir(bento.path)
                    except NotFound:
                        cloud_client = BentoMLContainer.bentocloud_client.get()
                        cloud_client.bento.pull(bento_tag, bento_store=tmp_bento_store)
                        return get_or_pull(bento_tag)

                return (get_or_pull, (self.bento.tag,))
        else:
            from bentoml._internal.service.loader import import_service

            return (import_service, self.get_service_import_origin())

    def __init__(
        self,
        name: str,
        *,
        runners: list[AbstractRunner] | None = None,
        models: list[Model] | None = None,
    ):
        """Service definition itself. Runners and models can be optionally pass into a ``bentoml.Service``.

        Args:
            name: name of the service
            runners: Optional list of runners to be used with this service.
            models: Optional list of ``bentoml.Model`` to be used with this service.
        """
        name = get_valid_service_name(name)

        # validate runners list contains Runner instances and runner names are unique
        if runners is not None:
            runner_names: t.Set[str] = set()
            for r in runners:
                assert isinstance(r, AbstractRunner), (
                    f'Service runners list can only contain bentoml.Runner instances, type "{type(r)}" found.'
                )

                if r.name in runner_names:
                    raise ValueError(
                        f"Found duplicate name `{r.name}` in service runners."
                    )
                runner_names.add(r.name)

        # validate models list contains Model instances
        if models is not None:
            for model in models:
                assert isinstance(model, Model), (
                    f'Service models list can only contain bentoml.Model instances, type "{type(model)}" found.'
                )

        self.__attrs_init__(  # type: ignore
            name=name,
            runners=[] if runners is None else runners,
            models=[] if models is None else models,
        )

        # Set import origin info - import_str can not be determined at this stage yet as
        # the variable name is only available in module vars after __init__ is returned
        # get_service_import_origin below will use the _caller_module for retriving the
        # correct import_str for this service
        caller_module = inspect.currentframe().f_back.f_globals["__name__"]
        self._caller_module = caller_module
        self._working_dir = os.getcwd()

    def get_service_import_origin(self) -> tuple[str, str]:
        """
        Returns the module name and working directory of the service
        """
        if not self._import_str:
            import_module = self._caller_module
            if import_module == "__main__":
                if hasattr(sys.modules["__main__"], "__file__"):
                    import_module = sys.modules["__main__"].__file__
                else:
                    raise BentoMLException(
                        "Failed to get service import origin, bentoml.Service object defined interactively in console or notebook is not supported"
                    )

            if self._caller_module not in sys.modules:
                raise BentoMLException(
                    "Failed to get service import origin, bentoml.Service object must be defined in a module"
                )

            for name, value in vars(sys.modules[self._caller_module]).items():
                if value is self:
                    object.__setattr__(self, "_import_str", f"{import_module}:{name}")
                    break

            if not self._import_str:
                raise BentoMLException(
                    "Failed to get service import origin, bentoml.Service object must be assigned to a variable at module level"
                )

        assert self._working_dir is not None

        return self._import_str, self._working_dir

    def is_service_importable(self) -> bool:
        if self._caller_module == "__main__":
            if not hasattr(sys.modules["__main__"], "__file__"):
                return False

        return True

    # fmt: off
    # case 1: function is not defined, but input and output are
    @t.overload
    def api(self, input: IODescriptor[IOType], output: IODescriptor[IOType]) ->  _inference_api_wrapper[IOType]: ...
    # case 2: the decorator itself with custom routes
    @t.overload
    def api(self, input: IODescriptor[IOType], output: IODescriptor[IOType], *, route: str = ...) ->  _inference_api_wrapper[IOType]: ...
    # fmt: on
    def api(
        self,
        input: IODescriptor[IOType],
        output: IODescriptor[IOType],
        *,
        name: str | None = None,
        doc: str | None = None,
        route: str | None = None,
    ) -> _inference_api_wrapper[IOType]:
        """Decorator for adding InferenceAPI to this service"""

        def decorator(
            fn: _inference_api_wrapper[IOType],
        ) -> _inference_api_wrapper[IOType]:
            _api = InferenceAPI[IOType](
                name=first_not_none(name, default=fn.__name__),
                user_defined_callback=fn,
                input_descriptor=input,
                output_descriptor=output,
                doc=doc,
                route=route,
            )
            if _api.name in self.apis:
                raise BentoMLException(
                    f"API {_api.name} is already defined in Service {self.name}"
                )
            self.apis[_api.name] = _api
            return fn

        return t.cast("_inference_api_wrapper[IOType]", decorator)

    def __repr__(self):
        if self.bento:
            return f'bentoml.Service(tag="{self.tag}", path="{self.bento.path}")'

        try:
            import_str, working_dir = self.get_service_import_origin()
            return (
                f'bentoml.Service(name="{self.name}", '
                f'import_str="{import_str}", '
                f'working_dir="{working_dir}")'
            )
        except BentoMLException:
            return (
                f'bentoml.Service(name="{self.name}", '
                f"runners=[{','.join([r.name for r in self.runners])}])"
            )

    def __str__(self) -> str:
        return repr(self)

    def __eq__(self, other: t.Any) -> bool:
        if not isinstance(other, Service):
            return NotImplemented
        if self is other:
            return True

        if self.bento and other.bento:
            return self.bento.tag == other.bento.tag

        try:
            return self.get_service_import_origin() == other.get_service_import_origin()
        except BentoMLException:
            return False

    @property
    def doc(self) -> str:
        if self.bento is not None:
            return self.bento.doc

        return get_default_svc_readme(self)

    @property
    def openapi_spec(self) -> OpenAPISpecification:
        from .openapi import generate_spec

        return generate_spec(self)

    def on_startup(self, func: HookF_ctx) -> HookF_ctx:
        self.startup_hooks.append(partial(func, self.context))
        return func

    def on_shutdown(self, func: HookF_ctx) -> HookF_ctx:
        self.shutdown_hooks.append(partial(func, self.context))
        return func

    def on_deployment(self, func: HookF) -> HookF:
        self.deployment_hooks.append(func)
        return func

    def on_asgi_app_startup(self) -> None:
        pass

    def on_asgi_app_shutdown(self) -> None:
        pass

    def on_grpc_server_startup(self) -> None:
        pass

    def on_grpc_server_shutdown(self) -> None:
        pass

    def get_grpc_servicer(
        self, protocol_version: str = LATEST_PROTOCOL_VERSION
    ) -> services.BentoServiceServicer:
        """
        Return a gRPC servicer instance for this service.

        Args:
            protocol_version: The protocol version to use for the gRPC servicer.

        Returns:
            A bento gRPC servicer implementation.
        """
        return importlib.import_module(
            f".grpc.servicer.{protocol_version}",
            package="bentoml._internal.server",
        ).create_bento_servicer(self)

    @property
    def grpc_servicer(self):
        return self.get_grpc_servicer(protocol_version=LATEST_PROTOCOL_VERSION)

    @property
    def asgi_app(self) -> "ext.ASGIApp":
        from ..server.http_app import HTTPAppFactory

        return HTTPAppFactory(self)()

    def mount_asgi_app(
        self, app: "ext.ASGIApp", path: str = "/", name: t.Optional[str] = None
    ) -> None:
        self.mount_apps.append((app, path, name))  # type: ignore

    def mount_wsgi_app(
        self, app: ext.WSGIApp, path: str = "/", name: t.Optional[str] = None
    ) -> None:
        from a2wsgi import WSGIMiddleware

        self.mount_apps.append((WSGIMiddleware(app), path, name))  # type: ignore

    def add_asgi_middleware(
        self, middleware_cls: t.Type[ext.AsgiMiddleware], **options: t.Any
    ) -> None:
        self.middlewares.append((middleware_cls, options))

    def mount_grpc_servicer(
        self,
        servicer_cls: ServicerClass,
        add_servicer_fn: AddServicerFn,
        service_names: list[str],
    ) -> None:
        self.mount_servicers.append((servicer_cls, add_servicer_fn, service_names))

    def add_grpc_interceptor(
        self, interceptor_cls: t.Type[grpc.aio.ServerInterceptor], **options: t.Any
    ) -> None:
        from bentoml.exceptions import BadInput

        if not issubclass(interceptor_cls, grpc.aio.ServerInterceptor):
            if isinstance(interceptor_cls, partial):
                if options:
                    logger.debug(
                        "'%s' is a partial class, hence '%s' will be ignored.",
                        interceptor_cls,
                        options,
                    )
                if not issubclass(interceptor_cls.func, grpc.aio.ServerInterceptor):
                    raise BadInput(
                        "'partial' class is not a subclass of 'grpc.aio.ServerInterceptor'."
                    )
                self.interceptors.append(interceptor_cls)
            else:
                raise BadInput(
                    f"{interceptor_cls} is not a subclass of 'grpc.aio.ServerInterceptor'."
                )

        self.interceptors.append(partial(interceptor_cls, **options))

    def add_grpc_handlers(self, handlers: list[grpc.GenericRpcHandler]) -> None:
        self.grpc_handlers.extend(handlers)

    def on_load_bento(self, bento: Bento) -> None:
        self.bento = bento



# type: ignore[reportMissingTypeStubs]
import typing as t
from typing import TYPE_CHECKING

from opentelemetry.sdk.trace.sampling import ALWAYS_ON
from opentelemetry.sdk.trace.sampling import Decision
from opentelemetry.sdk.trace.sampling import ParentBased
from opentelemetry.sdk.trace.sampling import SamplingResult
from opentelemetry.sdk.trace.sampling import StaticSampler
from opentelemetry.sdk.trace.sampling import TraceIdRatioBased
from opentelemetry.trace import get_current_span

if TYPE_CHECKING:
    from opentelemetry.trace import Context
    from opentelemetry.trace import Link
    from opentelemetry.trace import SpanKind
    from opentelemetry.trace import TraceState
    from opentelemetry.util.types import Attributes


def _get_parent_trace_state(parent_context: "Context") -> t.Optional["TraceState"]:
    parent_span_context = get_current_span(parent_context).get_span_context()
    if parent_span_context is None or not parent_span_context.is_valid:
        return None
    return parent_span_context.trace_state


class TraceIdRatioBasedAlwaysRecording(TraceIdRatioBased):
    """
    A trace Sampler that:
    * always recording (so that we can get the trace_id)
    * respect the parent's trace_state
    * ratio sampling
    """

    def should_sample(
        self,
        parent_context: t.Optional["Context"],
        trace_id: int,
        name: str,
        kind: t.Optional["SpanKind"] = None,
        attributes: t.Optional["Attributes"] = None,
        links: t.Optional[t.Sequence["Link"]] = None,
        trace_state: t.Optional["TraceState"] = None,
    ) -> "SamplingResult":
        decision = Decision.RECORD_ONLY
        if trace_id & self.TRACE_ID_LIMIT < self.bound:
            decision = Decision.RECORD_AND_SAMPLE
        if decision is Decision.RECORD_ONLY:
            pass
            # attributes = None
        return SamplingResult(
            decision,
            attributes,
            _get_parent_trace_state(parent_context),  # type: ignore[reportGeneralTypeIssues]
        )


class ParentBasedTraceIdRatio(ParentBased):
    """
    Sampler that respects its parent span's sampling decision, but otherwise
    samples probabalistically based on `rate`.
    """

    def __init__(self, rate: float):
        root = TraceIdRatioBasedAlwaysRecording(rate=rate)
        super().__init__(
            root=root,
            remote_parent_sampled=ALWAYS_ON,
            remote_parent_not_sampled=StaticSampler(Decision.RECORD_ONLY),
            local_parent_sampled=ALWAYS_ON,
            local_parent_not_sampled=StaticSampler(Decision.RECORD_ONLY),
        )



import sys
import warnings

import prometheus_client

warnings.warn(
    "bentoml.metrics module is deprecated and will be removed in the future. "
    "Please use prometheus_client directly for metrics reporting.",
    DeprecationWarning,
    stacklevel=1,
)

sys.modules[__name__] = prometheus_client



"""
User facing python APIs for deployment
"""

from __future__ import annotations

import typing as t

import attr
from simple_di import Provide
from simple_di import inject

from ._internal.cloud.deployment import Deployment
from ._internal.cloud.deployment import DeploymentConfigParameters
from ._internal.cloud.schemas.modelschemas import EnvItemSchema
from ._internal.cloud.schemas.modelschemas import LabelItemSchema
from ._internal.configuration.containers import BentoMLContainer
from ._internal.tag import Tag
from .exceptions import BentoMLException

if t.TYPE_CHECKING:
    from ._internal.cloud import BentoCloudClient


@t.overload
def create(
    name: str | None = ...,
    path_context: str | None = ...,
    *,
    bento: Tag | str | None = ...,
    cluster: str | None = ...,
    access_authorization: bool | None = ...,
    scaling_min: int | None = ...,
    scaling_max: int | None = ...,
    instance_type: str | None = ...,
    strategy: str | None = ...,
    envs: t.List[EnvItemSchema] | t.List[dict[str, t.Any]] | None = ...,
    extras: dict[str, t.Any] | None = ...,
) -> Deployment: ...


@t.overload
def create(
    name: str | None = ...,
    path_context: str | None = ...,
    *,
    bento: Tag | str | None = ...,
    config_file: str | None = ...,
) -> Deployment: ...


@t.overload
def create(
    name: str | None = ...,
    path_context: str | None = ...,
    *,
    bento: Tag | str | None = ...,
    config_dict: dict[str, t.Any] | None = ...,
) -> Deployment: ...


@inject
def create(
    name: str | None = None,
    path_context: str | None = None,
    *,
    bento: Tag | str | None = None,
    cluster: str | None = None,
    access_authorization: bool | None = None,
    scaling_min: int | None = None,
    scaling_max: int | None = None,
    instance_type: str | None = None,
    strategy: str | None = None,
    envs: t.List[EnvItemSchema] | t.List[dict[str, t.Any]] | None = None,
    extras: dict[str, t.Any] | None = None,
    config_dict: dict[str, t.Any] | None = None,
    config_file: str | None = None,
    _cloud_client: BentoCloudClient = Provide[BentoMLContainer.bentocloud_client],
) -> Deployment:
    config_params = DeploymentConfigParameters(
        name=name,
        path_context=path_context,
        bento=bento,
        cluster=cluster,
        access_authorization=access_authorization,
        scaling_max=scaling_max,
        scaling_min=scaling_min,
        instance_type=instance_type,
        strategy=strategy,
        envs=(
            [
                attr.asdict(item) if isinstance(item, EnvItemSchema) else item
                for item in envs
            ]
            if envs is not None
            else None
        ),
        extras=extras,
        config_dict=config_dict,
        config_file=config_file,
    )
    try:
        config_params.verify()
    except BentoMLException as e:
        raise BentoMLException(
            f"Failed to create deployment due to invalid configuration: {e}"
        )
    return _cloud_client.deployment.create(deployment_config_params=config_params)


@t.overload
def update(
    name: str | None = ...,
    path_context: str | None = ...,
    context: str | None = ...,
    cluster: str | None = ...,
    *,
    bento: Tag | str | None = ...,
    access_authorization: bool | None = ...,
    scaling_min: int | None = ...,
    scaling_max: int | None = ...,
    instance_type: str | None = ...,
    strategy: str | None = ...,
    envs: t.List[EnvItemSchema] | t.List[dict[str, t.Any]] | None = ...,
    secrets: t.List[str] | None = ...,
    extras: dict[str, t.Any] | None = ...,
) -> Deployment: ...


@t.overload
def update(
    name: str | None = ...,
    path_context: str | None = ...,
    context: str | None = ...,
    cluster: str | None = None,
    *,
    bento: Tag | str | None = ...,
    config_file: str | None = ...,
) -> Deployment: ...


@t.overload
def update(
    name: str | None = ...,
    path_context: str | None = ...,
    context: str | None = ...,
    cluster: str | None = None,
    *,
    bento: Tag | str | None = ...,
    config_dict: dict[str, t.Any] | None = ...,
) -> Deployment: ...


@inject
def update(
    name: str | None = None,
    path_context: str | None = None,
    cluster: str | None = None,
    *,
    bento: Tag | str | None = None,
    access_authorization: bool | None = None,
    scaling_min: int | None = None,
    scaling_max: int | None = None,
    instance_type: str | None = None,
    strategy: str | None = None,
    envs: (
        t.List[EnvItemSchema]
        | t.List[dict[str, t.Any]]
        | t.List[dict[str, t.Any]]
        | None
    ) = None,
    secrets: t.List[str] | None = None,
    extras: dict[str, t.Any] | None = None,
    config_dict: dict[str, t.Any] | None = None,
    config_file: str | None = None,
    _cloud_client: BentoCloudClient = Provide[BentoMLContainer.bentocloud_client],
) -> Deployment:
    config_params = DeploymentConfigParameters(
        name=name,
        path_context=path_context,
        bento=bento,
        cluster=cluster,
        access_authorization=access_authorization,
        scaling_max=scaling_max,
        scaling_min=scaling_min,
        instance_type=instance_type,
        strategy=strategy,
        envs=(
            [
                attr.asdict(item) if isinstance(item, EnvItemSchema) else item
                for item in envs
            ]
            if envs is not None
            else None
        ),
        secrets=secrets,
        extras=extras,
        config_dict=config_dict,
        config_file=config_file,
    )
    try:
        config_params.verify(create=False)
    except BentoMLException as e:
        raise BentoMLException(
            f"Failed to create deployment due to invalid configuration: {e}"
        )

    return _cloud_client.deployment.update(deployment_config_params=config_params)


@t.overload
def apply(
    name: str | None = ...,
    cluster: t.Optional[str] = ...,
    path_context: t.Optional[str] = ...,
    *,
    bento: t.Optional[t.Union[Tag, str]] = ...,
    config_dict: t.Optional[dict[str, t.Any]] = ...,
) -> Deployment: ...


@t.overload
def apply(
    name: str | None = ...,
    cluster: t.Optional[str] = ...,
    path_context: t.Optional[str] = ...,
    *,
    bento: t.Optional[t.Union[Tag, str]] = ...,
    config_file: t.Optional[str] = ...,
) -> Deployment: ...


@inject
def apply(
    name: str | None = None,
    cluster: str | None = None,
    path_context: str | None = None,
    *,
    bento: Tag | str | None = None,
    config_dict: dict[str, t.Any] | None = None,
    config_file: str | None = None,
    _cloud_client: BentoCloudClient = Provide[BentoMLContainer.bentocloud_client],
) -> Deployment:
    config_params = DeploymentConfigParameters(
        name=name,
        path_context=path_context,
        bento=bento,
        cluster=cluster,
        config_dict=config_dict,
        config_file=config_file,
    )
    try:
        config_params.verify(create=False)
    except BentoMLException as e:
        raise BentoMLException(
            f"Failed to create deployment due to invalid configuration: {e}"
        )

    return _cloud_client.deployment.apply(deployment_config_params=config_params)


@inject
def get(
    name: str,
    cluster: str | None = None,
    _cloud_client: BentoCloudClient = Provide[BentoMLContainer.bentocloud_client],
) -> Deployment:
    return _cloud_client.deployment.get(name=name, cluster=cluster)


@inject
def terminate(
    name: str,
    cluster: str | None = None,
    wait: bool = False,
    _cloud_client: BentoCloudClient = Provide[BentoMLContainer.bentocloud_client],
) -> Deployment:
    return _cloud_client.deployment.terminate(name=name, cluster=cluster, wait=wait)


@inject
def delete(
    name: str,
    cluster: str | None = None,
    _cloud_client: BentoCloudClient = Provide[BentoMLContainer.bentocloud_client],
) -> None:
    _cloud_client.deployment.delete(name=name, cluster=cluster)


@inject
def list(
    cluster: str | None = None,
    search: str | None = None,
    dev: bool = False,
    q: str | None = None,
    labels: t.List[LabelItemSchema] | t.List[dict[str, t.Any]] | None = None,
    _cloud_client: BentoCloudClient = Provide[BentoMLContainer.bentocloud_client],
) -> t.List[Deployment]:
    # Syntatic sugar to enable searching by `labels` argument
    if labels is not None:
        label_query = " ".join(
            f"label:{d['key']}={d['value']}"
            for d in (
                label if isinstance(label, dict) else attr.asdict(label)
                for label in labels
            )
        )

        if q is not None:
            q = f"{q} {label_query}"
        else:
            q = label_query

    return _cloud_client.deployment.list(cluster=cluster, search=search, dev=dev, q=q)


__all__ = ["create", "get", "update", "apply", "terminate", "delete", "list"]



from __future__ import annotations

import re
import typing as t
from copy import deepcopy
from numbers import Real

import schema as s

from ...utils.metrics import DEFAULT_BUCKET
from ...utils.unflatten import unflatten
from ..helpers import depth
from ..helpers import ensure_iterable_type
from ..helpers import ensure_larger_than
from ..helpers import ensure_larger_than_zero
from ..helpers import ensure_range
from ..helpers import is_valid_ip_address
from ..helpers import rename_fields
from ..helpers import validate_otlp_protocol
from ..helpers import validate_tracing_type

TRACING_CFG = {
    "exporter_type": s.Or(s.And(str, s.Use(str.lower), validate_tracing_type), None),
    "sample_rate": s.Or(s.And(float, ensure_range(0, 1)), None),
    "timeout": s.Or(s.And(int, ensure_larger_than_zero), None),
    "max_tag_value_length": s.Or(int, None),
    "excluded_urls": s.Or([str], str, None),
    "zipkin": {
        "endpoint": s.Or(str, None),
        "local_node_ipv4": s.Or(s.Or(s.And(str, is_valid_ip_address), int), None),
        "local_node_ipv6": s.Or(s.Or(s.And(str, is_valid_ip_address), int), None),
        "local_node_port": s.Or(s.And(int, ensure_larger_than_zero), None),
    },
    "jaeger": {
        "protocol": s.Or(
            s.And(str, s.Use(str.lower), lambda d: d in ["thrift", "grpc"]),
            None,
        ),
        "collector_endpoint": s.Or(str, None),
        "thrift": {
            "agent_host_name": s.Or(str, None),
            "agent_port": s.Or(int, None),
            "udp_split_oversized_batches": s.Or(bool, None),
        },
        "grpc": {
            "insecure": s.Or(bool, None),
        },
    },
    "otlp": {
        "protocol": s.Or(s.And(str, s.Use(str.lower), validate_otlp_protocol), None),
        "endpoint": s.Or(str, None),
        "compression": s.Or(
            s.And(str, lambda d: d in {"gzip", "none", "deflate"}), None
        ),
        "http": {
            "certificate_file": s.Or(str, None),
            "headers": s.Or(dict, None),
        },
        "grpc": {
            "insecure": s.Or(bool, None),
            "headers": s.Or(lambda d: isinstance(d, t.Sequence), None),
        },
    },
}
_API_SERVER_CONFIG = {
    "workers": s.Or(s.And(int, ensure_larger_than_zero), None),
    s.Optional("traffic"): {
        "timeout": s.And(Real, ensure_larger_than_zero),
        s.Optional("max_concurrency"): s.Or(s.And(int, ensure_larger_than_zero), None),
    },
    "backlog": s.And(int, ensure_larger_than(64)),
    "max_runner_connections": s.And(int, ensure_larger_than_zero),
    "metrics": {
        "enabled": bool,
        "namespace": str,
        s.Optional("duration"): {
            s.Optional("buckets", default=DEFAULT_BUCKET): s.Or(
                s.And(list, ensure_iterable_type(float)), None
            ),
            s.Optional("min"): s.Or(s.And(float, ensure_larger_than_zero), None),
            s.Optional("max"): s.Or(s.And(float, ensure_larger_than_zero), None),
            s.Optional("factor"): s.Or(s.And(float, ensure_larger_than(1.0)), None),
        },
    },
    "logging": {
        "access": {
            "enabled": bool,
            "request_content_length": s.Or(bool, None),
            "request_content_type": s.Or(bool, None),
            "response_content_length": s.Or(bool, None),
            "response_content_type": s.Or(bool, None),
            "format": {
                "trace_id": str,
                "span_id": str,
            },
        },
    },
    "http": {
        "host": s.And(str, is_valid_ip_address),
        "port": s.And(int, ensure_larger_than_zero),
        "cors": {
            "enabled": bool,
            "access_control_allow_origins": s.Or([str], str, None),
            "access_control_allow_origin_regex": s.Or(
                s.And(str, s.Use(re.compile)), None
            ),
            "access_control_allow_credentials": s.Or(bool, None),
            "access_control_allow_headers": s.Or([str], str, None),
            "access_control_allow_methods": s.Or([str], str, None),
            "access_control_max_age": s.Or(int, None),
            "access_control_expose_headers": s.Or([str], str, None),
        },
        "response": {"trace_id": bool},
    },
    "grpc": {
        "host": s.And(str, is_valid_ip_address),
        "port": s.And(int, ensure_larger_than_zero),
        "metrics": {
            "port": s.And(int, ensure_larger_than_zero),
            "host": s.And(str, is_valid_ip_address),
        },
        "reflection": {"enabled": bool},
        "channelz": {"enabled": bool},
        "max_concurrent_streams": s.Or(int, None),
        "max_message_length": s.Or(int, None),
        "maximum_concurrent_rpcs": s.Or(int, None),
    },
    s.Optional("ssl"): {
        "enabled": bool,
        s.Optional("certfile"): s.Or(str, None),
        s.Optional("keyfile"): s.Or(str, None),
        s.Optional("keyfile_password"): s.Or(str, None),
        s.Optional("version"): s.Or(s.And(int, ensure_larger_than_zero), None),
        s.Optional("cert_reqs"): s.Or(int, None),
        s.Optional("ca_certs"): s.Or(str, None),
        s.Optional("ciphers"): s.Or(str, None),
    },
    "runner_probe": {
        "enabled": bool,
        "timeout": int,
        "period": int,
    },
}
_RUNNER_CONFIG = {
    s.Optional("batching"): {
        s.Optional("enabled"): bool,
        s.Optional("max_batch_size"): s.And(int, ensure_larger_than_zero),
        s.Optional("max_latency_ms"): s.And(int, ensure_larger_than_zero),
    },
    # NOTE: there is a distinction between being unset and None here; if set to 'None'
    # in configuration for a specific runner, it will override the global configuration.
    s.Optional("resources"): s.Or(
        {s.Optional(str): object}, lambda s: s == "system", None
    ),  # type: ignore (incomplete schema typing)
    s.Optional("workers_per_resource"): s.And(
        s.Or(int, float), ensure_larger_than_zero
    ),
    s.Optional("logging"): {
        s.Optional("access"): {
            s.Optional("enabled"): bool,
            s.Optional("request_content_length"): s.Or(bool, None),
            s.Optional("request_content_type"): s.Or(bool, None),
            s.Optional("response_content_length"): s.Or(bool, None),
            s.Optional("response_content_type"): s.Or(bool, None),
        },
    },
    s.Optional("metrics"): {
        "enabled": bool,
        "namespace": str,
    },
    s.Optional("traffic"): {
        "timeout": s.And(Real, ensure_larger_than_zero),
        s.Optional("max_concurrency"): s.Or(s.And(int, ensure_larger_than_zero), None),
    },
}
SCHEMA = s.Schema(
    {
        s.Optional("version", default=1): s.And(int, lambda v: v == 1),
        "api_server": _API_SERVER_CONFIG,
        "runners": {
            **_RUNNER_CONFIG,
            s.Optional(str): _RUNNER_CONFIG,
        },
        "tracing": TRACING_CFG,
        s.Optional("monitoring"): {
            "enabled": bool,
            s.Optional("type"): s.Or(str, None),
            s.Optional("options"): s.Or(dict, None),
        },
        s.Optional("services"): s.Or(dict, None),
    }
)


def migration(*, override_config: dict[str, t.Any]):
    # We will use a flattened config to make it easier to migrate,
    # Then we will convert it back to a nested config.
    if depth(override_config) > 1:
        raise ValueError("'override_config' must be a flattened dictionary.") from None

    if "version" not in override_config:
        override_config["version"] = 1

    # First we migrate api_server field
    # 1. remove api_server.max_request_size (deprecated)
    rename_fields(
        override_config, current="api_server.max_request_size", remove_only=True
    )

    # 2. migrate api_server.[host|port] -> api_server.http.[host|port]
    for f in ["host", "port"]:
        rename_fields(
            override_config,
            current=f"api_server.{f}",
            replace_with=f"api_server.http.{f}",
        )

    # 3. migrate api_server.cors.[access_control_*] -> api_server.http.cors.[*]
    rename_fields(
        override_config,
        current="api_server.cors.enabled",
        replace_with="api_server.http.cors.enabled",
    )
    for f in [
        "allow_origin",
        "allow_credentials",
        "allow_headers",
        "allow_methods",
        "max_age",
        "expose_headers",
    ]:
        rename_fields(
            override_config,
            current=f"api_server.cors.access_control_{f}",
            replace_with=f"api_server.http.cors.access_control_{f}",
        )

    rename_fields(
        override_config,
        current="api_server.http.cors.access_control_allow_origin",
        replace_with="api_server.http.cors.access_control_allow_origins",
    )

    # 4. if ssl is present, in version 2 we introduce a api_server.ssl.enabled field to determine
    # whether user want to enable SSL.
    if len([f for f in override_config if f.startswith("api_server.ssl")]) != 0:
        override_config["api_server.ssl.enabled"] = True

    # 5. migrate all tracing fields to api_server.tracing
    # 5.1. migrate tracing.type -> api_server.tracing.expoerter_type
    rename_fields(
        override_config,
        current="tracing.type",
        replace_with="tracing.exporter_type",
    )
    # 5.2. for Zipkin and OTLP, migrate tracing.[exporter].url -> api_server.tracing.[exporter].endpoint
    for exporter in ["zipkin", "otlp"]:
        rename_fields(
            override_config,
            current=f"tracing.{exporter}.url",
            replace_with=f"tracing.{exporter}.endpoint",
        )
    # 5.3. For Jaeger, migrate tracing.jaeger.[address|port] -> api_server.tracing.jaeger.thrift.[agent_host_name|agent_port]
    rename_fields(
        override_config,
        current="tracing.jaeger.address",
        replace_with="tracing.jaeger.thrift.agent_host_name",
    )
    rename_fields(
        override_config,
        current="tracing.jaeger.port",
        replace_with="tracing.jaeger.thrift.agent_port",
    )
    # we also need to choose which protocol to use for jaeger.
    if (
        len(
            [
                f
                for f in override_config
                if f.startswith("api_server.tracing.jaeger.thrift")
            ]
        )
        != 0
    ):
        override_config["tracing.jaeger.protocol"] = "thrift"
    # 6. Last but not least, moving logging.formatting.* -> api_server.logging.access.format.*
    for f in ["trace_id", "span_id"]:
        rename_fields(
            override_config,
            current=f"logging.formatting.{f}_format",
            replace_with=f"api_server.logging.access.format.{f}",
        )
    # 7. move timeout to traffic.timeout
    for namespace in ("api_server", "runners"):
        rename_fields(
            override_config,
            current=f"{namespace}.timeout",
            replace_with=f"{namespace}.traffic.timeout",
        )
    for key in list(override_config):
        if key.startswith("runners."):
            runner_name = key.split(".")[1]
            if any(key.schema == runner_name for key in _RUNNER_CONFIG):
                continue
            rename_fields(
                override_config,
                current=f"runners.{runner_name}.timeout",
                replace_with=f"runners.{runner_name}.traffic.timeout",
            )

    return unflatten(override_config)


def finalize_config(config: dict[str, t.Any]) -> None:
    from ...utils import deep_merge

    # 8. if runner is overriden, set the runner default values
    RUNNER_CFG_KEYS = [
        "batching",
        "resources",
        "logging",
        "metrics",
        "traffic",
        "workers_per_resource",
    ]
    default_runner_config: dict[str, t.Any] = {
        key: value for key, value in config["runners"].items() if key in RUNNER_CFG_KEYS
    }

    for runner_name, runner_cfg in config["runners"].items():
        if runner_name in RUNNER_CFG_KEYS:
            continue
        # key is a runner name
        config["runners"][runner_name] = deep_merge(
            deepcopy(default_runner_config), runner_cfg
        )
